from typing import Annotated, Any, Generic, TypeVar

from smartspace.core import (
    Block,
    ChannelEvent,
    Config,
    InputChannel,
    OperatorBlock,
    Output,
    OutputChannel,
    State,
    Tool,
    callback,
    metadata,
    step,
)
from smartspace.enums import BlockCategory, ChannelState

ItemT = TypeVar("ItemT")
ResultT = TypeVar("ResultT")

@metadata(
    category=BlockCategory.FUNCTION,
    description="Loops through each item in the items input and sends them to the configured tool. Once all items have been processed, outputs the resulting list",
    icon="fa-project-diagram",
    label="map function, transform items, process list, iterate collection, apply function to list",
)
class Map(Block, Generic[ItemT, ResultT]):
    class Operation(Tool):
        def run(self, item: ItemT) -> ResultT: ...

    run: Operation

    results: Output[list[ResultT]]

    synchronous: Annotated[bool, Config()] = False

    items: Annotated[
        list[ItemT],
        State(
            step_id="map",
            input_ids=["items"],
        ),
    ] = []

    count: Annotated[
        int,
        State(
            step_id="map",
            input_ids=["items"],
        ),
    ] = 0

    results_state: Annotated[
        list[Any],
        State(
            step_id="map",
            input_ids=["items"],
        ),
    ] = []

    @step()
    async def map(self, items: list[ItemT]):
        if len(items) == 0:
            self.results.send([])
            return

        self.items = items
        self.results_state = [None] * len(items)
        self.count = len(items)
        if self.synchronous:
            await self.run.call(items[0]).then(lambda result: self.collect(result, 0))
        else:
            for i, item in enumerate(items):
                await self.run.call(item).then(lambda result: self.collect(result, i))

    @callback()
    async def collect(
        self,
        result: ResultT,
        index: int,
    ):
        self.results_state[index] = result
        self.count -= 1

        if self.count == 0:
            self.results.send(self.results_state)
        elif self.synchronous:
            i = len(self.items) - self.count
            await self.run.call(self.items[i]).then(
                lambda result: self.collect(result, i)
            )

@metadata(
    category=BlockCategory.FUNCTION,
    description="Collects data from a channel and outputs them as a list once the channel closes.",
    icon="fa-boxes",
    label="collect list, gather items, accumulate data, assemble collection, aggregate entries",
    obsolete=True,
)
class Collect(OperatorBlock, Generic[ItemT]):
    items: Output[list[ItemT]]

    items_state: Annotated[
        list[ItemT],
        State(
            step_id="collect",
            input_ids=["item"],
        ),
    ] = []

    @step()
    async def collect(
        self,
        item: InputChannel[ItemT],
    ):
        if (
            item.state == ChannelState.OPEN
            and item.event == ChannelEvent.DATA
            and item.data
        ):
            self.items_state.append(item.data)

        if item.event == ChannelEvent.CLOSE:
            self.items.send(self.items_state)

@metadata(
    category=BlockCategory.FUNCTION,
    description="Loops through a list of items and outputs them one at a time",
    icon="fa-ellipsis-h\t",
    label="for each, iterate items, loop through list, process each item, step through collection",
)
class ForEach(OperatorBlock, Generic[ItemT]):
    item: OutputChannel[ItemT]

    @step()
    async def foreach(self, items: list[ItemT]):
        for item in items:
            self.item.send(item)

        self.item.close()



from typing import Any, Generic, TypeVar

from smartspace.core import (
    Block,
    metadata,
    step,
)
from smartspace.enums import BlockCategory

SequenceT = TypeVar("SequenceT", bound=str | list[Any])


@metadata(
    category=BlockCategory.FUNCTION,
    description="Concatenates 2 lists or strings",
    icon="fa-plus",
    label="concatenate strings, join strings, merge lists, combine text, append strings",
)
class Concat(Block, Generic[SequenceT]):
    @step(output_name="result")
    async def concat(self, a: SequenceT, b: SequenceT) -> SequenceT:
        # Handle empty lists - don't add them
        if isinstance(a, list) and isinstance(b, list):
            if not a:  # a is empty
                return b
            if not b:  # b is empty
                return a
            # Both lists have content, concatenate normally
            return a + b  # type: ignore
        
        # For strings or mixed types, use normal concatenation
        return a + b  # type: ignore



from typing import Any, Generic, TypeVar

from pydantic import BaseModel, ConfigDict, ValidationError

from smartspace.core import (
    Block,
    GenericSchema,
    Output,
    metadata,
    step,
)
from smartspace.enums import BlockCategory

ItemT = TypeVar("ItemT")


class TypeSwitchOption(Generic[ItemT]):
    schema: GenericSchema[ItemT]
    option: Output[ItemT]


@metadata(
    description="Checks the output schemas of the options and sends the input to the first option that matches",
    category=BlockCategory.MISC,
    icon="fa-random",
    label="type switch, schema routing, type routing, data branching, conditional path",
)
class TypeSwitch(Block):
    options: list[TypeSwitchOption]

    @step(output_name="result")
    async def switch(self, item: Any):
        for option in self.options:

            class M(BaseModel):
                model_config = ConfigDict(
                    json_schema_extra=option.schema,
                )

            try:
                option.option.send(M.model_validate(item, strict=True))
            except ValidationError:
                ...



import asyncio
import json
from typing import List

import pydantic_core
import requests
import typer
from pydantic import TypeAdapter

from smartspace.cli.models import PublishedBlockSet
from smartspace.core import BlockSet
from smartspace.models import (
    BlockRunData,
)

app = typer.Typer()


def get_config():
    import smartspace.cli.auth
    import smartspace.cli.config

    config = smartspace.cli.config.load_config()

    if not config["config_api_url"]:
        print(
            "You must set your API url before using the CLI. Use 'smartspace config --api-url <Your SmartSpace API Url>'"
        )
        exit()

    return config


@app.command()
def list():
    import smartspace.blocks
    import smartspace.cli.auth

    config = get_config()

    response = requests.get(
        url=f"{config['config_api_url']}/blocksets",
        params={"types": "Custom"},
        headers={"Authorization": f"Bearer {smartspace.cli.auth.get_token()}"},
    )

    if response.status_code == 200:
        type_adapter = TypeAdapter(List[PublishedBlockSet])
        block_sets = type_adapter.validate_json(response.content)
        for block_set in block_sets:
            print(f"{block_set.name}:")
            for block_name, versions in block_set.block_interfaces.items():
                for version in versions.keys():
                    print(f"  {block_name} ({version})")
    else:
        print(f"Error: {response.text}")


@app.command()
def delete(name: str):
    import smartspace.blocks
    import smartspace.cli.auth

    config = get_config()

    response = requests.delete(
        url=f"{config['config_api_url']}/blocksets/{name}",
        headers={"Authorization": f"Bearer {smartspace.cli.auth.get_token()}"},
    )

    if response.status_code >= 200 and response.status_code < 300:
        print("Successfully deleted")
    elif response.status_code == 500:
        print("An internal error occured while trying to delete the block set")
    else:
        result = json.loads(response.content)
        print(result["detail"] if "detail" in result else json.dumps(result, indent=2))


@app.command()
def publish(name: str, path: str = ""):
    import os
    import zipfile

    import smartspace.blocks
    import smartspace.cli.auth

    config = get_config()

    file_name = "blocks.zip"
    if os.path.exists(file_name):
        os.remove(file_name)

    block_set = asyncio.run(smartspace.blocks.load(path, force_reload=True))

    print("Publishing the following blocks:")
    for block_name, versions in block_set.all.items():
        for version, _ in versions.items():
            print(f"{block_name} ({version})")

    zf = zipfile.ZipFile(file_name, "w")
    for dirname, subdirs, files in os.walk(path):
        zf.write(dirname)
        for filename in files:
            zf.write(os.path.join(dirname, filename))
    zf.close()

    with open(file_name, "rb") as f:
        response = requests.post(
            url=f"{config['config_api_url']}/blocksets/{name}",
            headers={"Authorization": f"Bearer {smartspace.cli.auth.get_token()}"},
            files={"file": f},
        )
        if response.status_code == 200:
            print("Published!")
        else:
            print(f"Error: {response.text}")

    if os.path.exists(file_name):
        os.remove(file_name)


@app.command()
def debug(path: str = "", poll: bool = False):
    import asyncio
    import os
    from contextlib import suppress

    from pysignalr.client import SignalRClient
    from pysignalr.messages import (
        CompletionMessage,
        HandshakeMessage,
        InvocationMessage,
        Message,
    )
    from pysignalr.protocol.json import JSONProtocol, MessageEncoder
    from watchdog.events import FileSystemEvent, FileSystemEventHandler
    from watchdog.observers import Observer
    from watchdog.observers.polling import PollingObserver

    import smartspace.blocks
    import smartspace.cli.auth

    config = get_config()

    root_path = path if path != "" else os.getcwd()

    print(f"Debugging blocks in '{root_path}'")

    message_encoder = MessageEncoder()

    class MyJSONProtocol(JSONProtocol):
        def encode(self, message: Message | HandshakeMessage) -> str:
            if isinstance(message, CompletionMessage):
                data = message.dump()
                if "error" in data and data["error"] is None:
                    del data["error"]
                return message_encoder.encode(data)
            else:
                return JSONProtocol.encode(self, message)

    client = SignalRClient(
        url=f"{config['config_api_url']}debug"
        if config["config_api_url"].endswith("/")
        else f"{config['config_api_url']}/debug",
        headers={"Authorization": f"Bearer {smartspace.cli.auth.get_token()}"},
        protocol=MyJSONProtocol(),
    )

    block_set: BlockSet = BlockSet()

    running = (
        asyncio.Lock()
    )  # temp fix to deal with server issue when running blocks in parallel

    async def on_message_override(message: Message):
        if isinstance(message, InvocationMessage) and message.target == "run_block":
            async with running:
                request = BlockRunData.model_validate(message.arguments[0])

                print(
                    f"Running '{request.name}({request.version}).{request.function}()'"
                )

                block_type = block_set.find(request.name, request.version)

                if not block_type:
                    raise Exception(
                        f"Could not find block with name {request.name} and version {request.version}"
                    )

                block_instance = block_type()

                block_instance._load(
                    context=request.context,
                    state=request.state,
                    inputs=request.inputs,
                    dynamic_ports=request.dynamic_ports,
                    dynamic_output_pins=request.dynamic_output_pins,
                    dynamic_input_pins=request.dynamic_input_pins,
                )

                messages: List[dict] = []

                async for m in await block_instance._run_function(request.function):
                    messages.append(m.model_dump(by_alias=True, mode="json"))

                invocation_id = getattr(message, "invocation_id", None) or getattr(
                    message, "invocationId", ""
                )

                message = CompletionMessage(
                    invocation_id,
                    messages,
                    headers=client._headers,
                )
                print(
                    f"Finished '{request.name}({request.version}).{request.function}()'"
                )
                await client._transport.send(message)
                await asyncio.sleep(5)
        else:
            await SignalRClient._on_message(client, message)

    client._on_message = on_message_override
    client._transport._callback = on_message_override

    async def on_close() -> None:
        print("Disconnected from the server")

    async def on_error(message: CompletionMessage) -> None:
        print(f"Received error: {message.error}")

    async def on_open() -> None:
        await register_blocks(root_path)

    async def register_blocks(path: str):
        nonlocal block_set

        new_block_set = await smartspace.blocks.load(path, force_reload=True)
        old_blocks = block_set.all

        found_blocks = {
            block_name: {
                version: block_type.interface()
                for version, block_type in versions.items()
            }
            for block_name, versions in new_block_set.all.items()
        }

        new_blocks = {
            found_block_name: {
                version: block_interface
                for version, block_interface in found_block_versions.items()
                if not any(
                    [
                        old_block_name == found_block_name and version in old_versions
                        for old_block_name, old_versions in old_blocks.items()
                    ]
                )
            }
            for found_block_name, found_block_versions in found_blocks.items()
        }

        removed_blocks = {
            old_block_name: [
                old_version
                for old_version in old_versions.keys()
                if old_block_name not in found_blocks
                or old_version not in found_blocks[old_block_name]
            ]
            for old_block_name, old_versions in old_blocks.items()
        }

        updated_blocks = {
            found_block_name: {
                version: block_interface
                for version, block_interface in found_block_versions.items()
                if any(
                    [
                        old_block_name == found_block_name
                        and version in old_versions
                        and block_interface != old_versions[version].interface()
                        for old_block_name, old_versions in old_blocks.items()
                    ]
                )
            }
            for found_block_name, found_block_versions in found_blocks.items()
        }

        has_updated_blocks = False

        for block_name, versions in updated_blocks.items():
            for version, interface in versions.items():
                has_updated_blocks = True
                print(f"Updating {block_name} ({version})")

        if has_updated_blocks:
            data = pydantic_core.to_jsonable_python(updated_blocks)
            await client.send("registerblock", [data])

        has_new_blocks = False

        for block_name, versions in new_blocks.items():
            for version, interface in versions.items():
                has_new_blocks = True
                print(f"Registering {block_name} ({version})")

        if has_new_blocks:
            data = pydantic_core.to_jsonable_python(new_blocks)
            await client.send("registerblock", [data])

        for block_name, removed_versions in removed_blocks.items():
            for version in removed_versions:
                print(f"Removing {block_name} ({version})")
                await client.send(
                    "removeblock", [{"name": block_name, "version": version}]
                )

        if not len(new_block_set.all):
            print("Found no blocks")

        block_set = new_block_set

    client.on_open(on_open)
    client.on_close(on_close)
    client.on_error(on_error)

    class _EventHandler(FileSystemEventHandler):
        def __init__(self, loop: asyncio.AbstractEventLoop, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.loop = loop

        def _on_any_event(self, event: FileSystemEvent):
            asyncio.run_coroutine_threadsafe(register_blocks(root_path), self.loop)

        def on_created(self, event: FileSystemEvent):
            self._on_any_event(event)

        def on_deleted(self, event: FileSystemEvent):
            self._on_any_event(event)

        def on_modified(self, event: FileSystemEvent):
            self._on_any_event(event)

        def on_moved(self, event: FileSystemEvent):
            self._on_any_event(event)

    async def main():
        loop = asyncio.get_event_loop()
        handler = _EventHandler(loop)
        observer = PollingObserver() if poll else Observer()
        observer.schedule(handler, root_path, recursive=True)
        observer.start()

        await client.run()

    with suppress(KeyboardInterrupt, asyncio.CancelledError):
        asyncio.run(main())


if __name__ == "__main__":
    debug()



from typing import Annotated, Any, Generic, TypeVar

from more_itertools import flatten

from smartspace.core import (
    Block,
    Config,
    OperatorBlock,
    Output,
    State,
    metadata,
    step,
)
from smartspace.enums import BlockCategory

ItemT = TypeVar("ItemT")
ResultT = TypeVar("ResultT")
SequenceT = TypeVar("SequenceT", bound=list[Any] | str)
firstItemT = TypeVar("firstItemT")


@metadata(
    category=BlockCategory.FUNCTION,
    icon="fa-sort-numeric-up",
    label="count items, list length, item count, size of list, total elements",
)
class Count(OperatorBlock):
    @step(output_name="output")
    async def count(self, items: list[Any]) -> int:
        return len(items)


@metadata(
    category=BlockCategory.FUNCTION,
    description="Joins a list of strings using the configured separator and outputs the resulting string.",
    icon="fa-link",
    label="join strings, concatenate text, combine strings, merge text, connect strings",
)
class JoinStrings(Block):
    separator: Annotated[str, Config()] = ""

    @step(output_name="output")
    async def join(self, strings: list[str]) -> str:
        return self.separator.join(strings)


@metadata(
    category=BlockCategory.FUNCTION,
    description="Splits a string using the configured separator and outputs a list of the substrings",
    icon="fa-cut",
    label="split string, divide text, break string, tokenize text, parse string",
)
class SplitString(Block):
    separator: Annotated[str, Config()] = "\n"
    include_separator: Annotated[bool, Config()] = False

    @step(output_name="output")
    async def split(self, string: str) -> list[str]:
        results = string.split(self.separator)

        if self.include_separator:
            results = [r + self.separator for r in results[:-1]] + [results[-1]]

        return results


@metadata(
    category=BlockCategory.FUNCTION,
    description="Slices a list or string using the configured start and end indexes.",
    icon="fa-cut",
    label="slice list, extract portion, get segment, subset sequence, partial list",
)
class Slice(Block):
    start: Annotated[int, Config()] = 0
    end: Annotated[int, Config()] = 0

    @step(output_name="items")
    async def slice(self, items: list[Any] | str) -> list[Any] | str:
        return items[self.start : self.end]


@metadata(
    category=BlockCategory.FUNCTION,
    description="Gets the first item from a list",
    icon="fa-arrow-alt-circle-left",
)
class First(OperatorBlock, Generic[firstItemT]):
    @step(output_name="item")
    async def first(self, items: list[firstItemT]) -> firstItemT:
        return items[0]


@metadata(
    category=BlockCategory.FUNCTION,
    description="Flattens a list of lists into a single list",
    icon="fa-compress",
    label="flatten list, merge nested lists, combine nested arrays, unnest lists, simplify nested lists",
)
class Flatten(OperatorBlock):
    @step(output_name="list")
    async def flatten(self, lists: list[list[Any]]) -> list[Any]:
        return list(flatten(lists))


@metadata(
    description="Takes in inputs and creates a list containing the inputs.",
    category=BlockCategory.MISC,
    icon="fa-list-ul",
    label="create list, build list, construct list, form list, make list",
)
class CreateList(Block):
    @step(output_name="list")
    async def build(self, *items: Any) -> list[Any]:
        return list(items)


@metadata(
    category=BlockCategory.FUNCTION,
    description="Appends an item to a list and outputs the updated list. Maintains the list state across calls.",
    label="list builder, dynamic list, item aggregation, list accumulation, append to list",
)
class BuildList(Block):
    items: Annotated[list[Any], State()] = []

    @step(output_name="items")
    async def create_response(self, item: Any) -> list[Any]:
        self.items.append(item)
        return self.items


@metadata(
    category=BlockCategory.FUNCTION,
    description="Merges objects from two lists by matching on the configured key.",
    obsolete=True,
    label="merge JSON lists, combine JSON arrays, join JSON objects, match and merge, consolidate JSON data",
    use_instead="Join",
)
class MergeLists(Block):
    key: Annotated[str, Config()]

    @step(output_name="result")
    async def merge_lists(
        self,
        a: list[dict[str, Any]],
        b: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        dict1 = {item[self.key]: item for item in a}
        dict2 = {item[self.key]: item for item in b}

        merged_dict = {}
        for code in dict1.keys() | dict2.keys():
            if code in dict1 and code in dict2:
                merged_dict[code] = {**dict1[code], **dict2[code]}
            elif code in dict1:
                merged_dict[code] = dict1[code]
            elif code in dict2:
                merged_dict[code] = dict2[code]

        final_result = list(merged_dict.values())

        return final_result


@metadata(
    description="Takes in a list and sends each item to the corresponding output",
    category=BlockCategory.MISC,
    icon="fa-th-list",
    label="unpack list, distribute items, extract list elements, spread array, decompose list",
)
class UnpackList(Block):
    items: list[Output[Any]]

    @step()
    async def unpack(self, list: list[Any]):
        for i, v in enumerate(list):
            if len(self.items) > i:
                self.items[i].send(v)


@metadata(
    description="Appends item to items and output resulting list",
    category=BlockCategory.MISC,
    icon="fa-plus",
    label="append item, add item, extend list, insert item, concatenate item",
)
class Append(Block, Generic[ItemT]):
    @step(output_name="items")
    async def build(self, items: list[ItemT], item: ItemT) -> list[ItemT]:
        items.append(item)
        return items



import pytest

from smartspace.blocks.website_scraper import WebsiteScraper


@pytest.mark.asyncio
async def test_make_request_get():
    w = WebsiteScraper()
    w.page_limit = 1
    await w.scrape_website("https://google.com")
    messages = w.get_messages()
    assert len(messages[0].outputs[0].value) == 1  # 1 message should be sent
    assert (
        "Google" in messages[0].outputs[0].value[0]
    )  # Google should be in the scraped content


@pytest.mark.asyncio
async def test_make_request_error():
    w = WebsiteScraper()
    w.page_limit = 1
    await w.scrape_website("https://dummyjson.com/http/404/Hello_Peter")
    messages = w.get_messages()
    assert len(messages[0].outputs[0].value) == 1
    assert messages[0].outputs[0].value[0] == ""


@pytest.mark.asyncio
async def test_make_request_multiple_pages():
    w = WebsiteScraper()
    page_limit = 3
    w.page_limit = page_limit
    await w.scrape_website("https://google.com")
    messages = w.get_messages()
    assert len(messages[0].outputs[0].value) == page_limit
    assert any("Google" in output.value[0] for output in messages[0].outputs)


@pytest.mark.asyncio
async def test_make_request_no_url():
    w = WebsiteScraper()
    w.page_limit = 1
    await w.scrape_website("")
    messages = w.get_messages()
    assert len(messages[0].outputs[0].value) == 1
    assert messages[0].outputs[0].value[0] == ""


@pytest.mark.asyncio
async def test_make_request_invalid_url():
    w = WebsiteScraper()
    w.page_limit = 1
    await w.scrape_website("invalid_url")
    messages = w.get_messages()
    assert len(messages[0].outputs[0].value) == 1
    assert messages[0].outputs[0].value[0] == ""



from typing import TypedDict


class SmartSpaceConfig(TypedDict):
    client_id: str | None
    tenant_id: str | None
    config_api_url: str | None


def load_config() -> SmartSpaceConfig:
    import json

    try:
        with open(_get_config_file_path(), "r") as f:
            return json.load(f)
    except FileNotFoundError:
        return SmartSpaceConfig(client_id=None, config_api_url=None, tenant_id=None)


def save_config(config: SmartSpaceConfig):
    import json
    import os
    from pathlib import Path

    Path(os.path.join(os.path.expanduser("~"), ".smartspace")).mkdir(
        parents=True, exist_ok=True
    )

    with open(_get_config_file_path(), "w+") as f:
        json.dump(config, f)


def _get_config_file_path():
    import os

    home_directory = os.path.expanduser("~")
    return os.path.join(home_directory, ".smartspace", "config.json")



from typing import Any

from lark import Lark, Token, Transformer

#
# ------------------- GRAMMAR -------------------
#
grammar = r"""
VALUE.2: /(?i)value/
LEN.2:   /(?i)len/

LPAR: "("
RPAR: ")"
DOT: "."

%import common.CNAME -> BASE_CNAME
CNAME: BASE_CNAME

%import common.INT
%import common.FLOAT
%import common.WS
%ignore WS

NUMBER: INT | FLOAT

STRING: /'[^']*'/ | /"[^"]*"/

NONE: /(?i)(none|null)/
TRUE: /(?i)true/
FALSE: /(?i)false/

?literal: list_literal
        | dict_literal

?list_literal: "[" [expr_list] "]" | "[]"
?expr_list: expr ("," expr)*

?dict_literal: "{" [pair_list] "}" | "{}"
?pair_list: pair ("," pair)*
?pair: STRING ":" expr

?start: expr

?expr:
    | expr "and"i expr               -> and_
    | expr "or"i expr                -> or_
    | expr OPERATOR expr             -> compare
    | expr "is"i "not"i NONE         -> is_not_none
    | expr "is"i "not"i FALSE        -> is_not_false
    | expr "in"i expr                -> in_
    | expr "not"i "in"i expr         -> not_in_
    | LEN LPAR expr RPAR OPERATOR NUMBER -> length_compare
    | value_attr
    | NUMBER                         -> number
    | STRING                         -> string
    | NONE                           -> none_
    | TRUE                           -> true_
    | FALSE                          -> false_
    | dict_literal                          -> dict_literal
    | list_literal                          -> list_literal
    | LPAR expr RPAR                 -> parens

OPERATOR: "==" | "!=" | "<>" | ">" | "<" | ">=" | "<=" | "~="

?value_attr: VALUE property_list
?property_list: (dot_property | bracket_index)*

dot_property: DOT CNAME
bracket_index: "[" NUMBER "]"
"""


class ConditionEvaluator(Transformer):
    """
    Walks the parse tree from the grammar and evaluates expressions
    against the provided `context` dict (or any Python object).
    """

    def __init__(self, context: Any):
        super().__init__()
        self.context = context  # The "value" in expressions

    #
    # ----- Logical operators -----
    #
    def and_(self, args):
        return args[0] and args[1]

    def or_(self, args):
        return args[0] or args[1]

    #
    # ----- "is not none", "is not false" -----
    #
    def is_not_none(self, args):
        # e.g. value.user.name is not none
        return args[0] is not None

    def is_not_false(self, args):
        # e.g. value.user.profile.active is not false
        return args[0] is not False

    #
    # ----- "in" / "not in" -----
    #
    def in_(self, args):
        """
        expr "in" expr
        e.g. value in [10, 15, 20]
             value.user.name in "Alice"
        """
        left, right = args
        return self._check_membership(left, right)

    def not_in_(self, args):
        left, right = args
        return not self._check_membership(left, right)

    def _check_membership(self, left, container):
        """
        Implement "left in container" for list, dict, str, etc.
        """
        if isinstance(container, list):
            container = [i.value if isinstance(i, Token) else i for i in container]

        if isinstance(container, dict):
            container = {
                k.value if isinstance(k, Token) else k: v.value
                if isinstance(v, Token)
                else v
                for k, v in container.items()
            }

        if isinstance(container, dict):
            return left in container
        elif isinstance(container, (list, str)):
            return left in container
        # You can decide how to handle other types:
        return False

    #
    # ----- len(...) OP NUMBER -----
    #
    def length_compare(self, args):
        """
        e.g. len(value.items) == 3
        """
        # parse layout: [Token(LEN), Token(LPAR), expr, Token(RPAR), operator, number]
        container, operator, number = args[2], args[4], args[5]

        # Convert operator, number from Token if needed
        if isinstance(operator, Token):
            operator = operator.value
        if isinstance(number, Token):
            number = float(number.value)

        if not isinstance(number, float):
            number = float(number)

        # container could be str, list, dict
        if not isinstance(container, (list, dict, str)):
            return False

        left_side = len(container)
        return self.apply_operator(left_side, operator, number)

    #
    # ----- Comparisons (expr OPERATOR expr) -----
    #
    def compare(self, args):
        """
        e.g. value.user.age > 18
             "abc" ~= "b"
             'foo' != 'bar'
             value.items[0] <= 10
             value == {"test": 1}
        """
        left, operator, right = args
        return self.apply_operator(left, operator, right)

    #
    # ----- Parentheses -----
    #
    def parens(self, args):
        return args[0]

    #
    # ----- value_attr (like value.user.age, etc.) -----
    #
    def value_attr(self, children):
        """
        children might be [Token('VALUE','value'), property_list([...])]
        If no property_list, it's just 'value'.
        """
        if len(children) == 1:
            # means there's no property_list, so it's just 'value'
            return self.context
        else:
            path_list = children[1]  # output from property_list
            return self._get_nested(path_list)

    def property_list(self, children):
        """
        property_list is a sequence of dot_property or bracket_index subtrees,
        each returning something like ["user"] or ["2"].
        We'll flatten them into one path list, e.g. ["user","age"].
        """
        path = []
        for child in children:
            path.extend(child)
        return path

    def dot_property(self, tokens):
        # tokens = [Token('DOT','.'), Token('CNAME','user')]
        return [tokens[1].value]

    def bracket_index(self, tokens):
        # tokens = [Token('[','['), Token('NUMBER','2'), Token(']',']')]
        if len(tokens) > 1:
            return [tokens[1].value]
        else:
            return [tokens[0].value]

    def _get_nested(self, path_list):
        """
        Actually retrieve the nested object from self.context,
        e.g. path_list = ["user","age"] or ["items","2"].
        """
        obj = self.context
        for part in path_list:
            if isinstance(obj, dict):
                obj = obj.get(part)
            elif isinstance(obj, list) and part.isdigit():
                idx = int(part)
                if 0 <= idx < len(obj):
                    obj = obj[idx]
                else:
                    obj = None
            else:
                # last resort: attribute access
                obj = getattr(obj, part, None)
        return obj

    #
    # ----- Literal lists & dicts -----
    #
    def list_literal(self, children):
        """
        list_literal: "[" [expr_list] "]"
        children might be empty (no expr_list) or one child which is expr_list
        """
        if not children:
            return []  # empty list, e.g. []
        else:
            # children[0] is the result of expr_list
            if isinstance(children[0], list) and len(children) == 1:
                return children[0]
            return children

    def expr_list(self, children):
        """
        expr_list: expr ("," expr)* -> a list of expressions
        """
        return children  # e.g. [expr1, expr2, expr3, ...]

    def dict_literal(self, children):
        """
        dict_literal: "{" [pair_list] "}"
        children might be empty or one child which is pair_list
        """
        if not children:
            return {}
        else:
            d = {}
            pairs = children if isinstance(children, list) else children.items()
            for pair in pairs:
                # each pair is (key, value)
                if isinstance(pair, dict):
                    d.update(pair)
                else:
                    k, v = pair
                    d[k] = v
            return d
            # children[0] is the pair_list

    def pair_list(self, children):
        """
        pair_list: pair ("," pair)*
        We'll merge them into a single dict
        """
        d = {}
        for pair in children:
            # each pair is (key, value)
            k, v = pair
            d[k] = v
        return d

    def pair(self, children):
        """
        pair: string ":" expr
        So children = [string_value, expr_value]
        """
        key = self.string(
            [children[0].value if isinstance(children[0], Token) else children[0]]
        )
        val = children[1].value if isinstance(children[1], Token) else children[1]
        return (key, val)

    #
    # ----- Terminal tokens: string, none, true, false, number -----
    #
    def string(self, args):
        """
        Convert e.g. "'Ali'" -> "Ali"
        """
        s = args[0]
        if (s.startswith("'") and s.endswith("'")) or (
            s.startswith('"') and s.endswith('"')
        ):
            return s[1:-1]
        return s

    def none_(self, _):
        return None

    def true_(self, _):
        return True

    def false_(self, _):
        return False

    def number(self, args):
        raw = args[0]
        if "." in raw:
            return float(raw)
        return int(raw)

    #
    # ----- Core operator logic -----
    #
    def apply_operator(self, left, operator, right):
        """
        We have standard operators + "<>" for != + "~=" for substring match.
        """
        # If either side is None, turn it into "" for string comparisons:
        if left is None:
            left = ""
        if right is None:
            right = ""

        # "~=" => substring check: str(left) contains str(right)
        if operator == "~=":
            return str(right) in str(left)

        # If it's == or != and either side is a string, do str compare
        if operator in ("==", "!=", "<>"):
            if isinstance(left, str) or isinstance(right, str):
                left = str(left)
                right = str(right)

        # Numeric comparisons
        if operator in (">", "<", ">=", "<=", "!=", "==", "<>"):
            # If either side is numeric, attempt float conversion
            if any(isinstance(x, (int, float)) for x in (left, right)):
                try:
                    left = float(left)
                    right = float(right)
                except ValueError:
                    return False

        # Define standard operators
        ops = {
            "==": lambda a, b: a == b,
            "!=": lambda a, b: a != b,
            "<>": lambda a, b: a != b,  # alternate not-equal
            ">": lambda a, b: a > b,
            "<": lambda a, b: a < b,
            ">=": lambda a, b: a >= b,
            "<=": lambda a, b: a <= b,
        }
        return ops[operator](left, right)


_parser = None


def evaluate_expression(condition: str, value: Any = None):
    """
    Main entry point: parse the condition string,
    then evaluate it against the given value or dict context.
    """
    global _parser
    if _parser is None:
        evaluate_expression = Lark(grammar, parser="lalr")
    tree = evaluate_expression.parse(condition)
    return ConditionEvaluator(value).transform(tree)


expression_tooltip = """Use value to reference your data (dict/list/etc.), e.g. value.user.age > 18.

Supports and, or, and operators like ==, !=, <, >, >=, <=, ~= (substring).

Check nested fields (value.user.profile.active is not false), lengths (len(value.items) == 3), and membership ('read' in value.user.permissions).

Literal lists/dicts let you do value in [10, null, "foo"] or value == {"test": 1}.

none|null|true|false are parsed as Python None|None|True|False"""

#
# ------------------- EXAMPLE & TESTS -------------------
#
if __name__ == "__main__":
    example_data = {
        "user": {
            "name": "Alice",
            "age": 25,
            "permissions": ["read", "write", "admin"],
            "profile": {"active": True},
        },
        "items": [10, 20, 30],
    }
    assert evaluate_expression("value == [{'text': 'test'}]", [{"text": "test"}])
    assert evaluate_expression("value in [10, 15, 20]", 10)  # => True

    # Basic tests
    assert evaluate_expression("value is not none", example_data)
    assert evaluate_expression("value.user.age > 18", example_data)
    assert evaluate_expression("value.user.name is not none", example_data)
    assert evaluate_expression("'read' in value.user.permissions", example_data)
    assert evaluate_expression("'delete' not in value.user.permissions", example_data)
    assert evaluate_expression("value.items[0] == 10", example_data)
    assert evaluate_expression("value.items[1] > 15", example_data)
    assert evaluate_expression("value.user.profile.active is not False", example_data)
    assert evaluate_expression("len(value.user.permissions) > 2", example_data)
    assert evaluate_expression("len(value.items) == 3", example_data)
    assert evaluate_expression("value.items[2] <= 30", example_data)
    assert evaluate_expression("value.user.age != 30", example_data)
    assert not evaluate_expression("value.user.age < 20", example_data)
    assert not evaluate_expression("value.user.nonexistent is not none", example_data)
    assert not evaluate_expression("value.items[100] > 10", example_data)

    # Alternate not-equal "<>"
    assert evaluate_expression("value.user.age <> 26", example_data)  # 25 != 26 => True

    # Substring operator "~="
    assert evaluate_expression(
        'value.user.name ~= "Ali"', example_data
    )  # "Alice" contains "Ali"
    assert not evaluate_expression('value.user.name ~= "icex"', example_data)

    # none/null/true/false
    assert evaluate_expression("value.user.age != none", example_data)
    assert evaluate_expression("value.user.age != null", example_data)
    assert evaluate_expression("value.user.profile.active == true", example_data)
    assert not evaluate_expression("value.user.profile.active == false", example_data)

    # "value.nobody" => None => == none => True
    assert evaluate_expression("value.nobody == none", example_data)

    # Quoted strings are stripped
    assert evaluate_expression('"Ali" == "Ali"', example_data)
    assert not evaluate_expression('"Ali" == "alice"', example_data)

    #
    # ---------- NEW TESTS for list/dict literal ----------
    #

    # 1) Membership in a literal list
    assert evaluate_expression("value in [10, 15, 20]", 10)  # => True
    assert not evaluate_expression("value in [10, 15, 20]", 5)  # => False
    # Mixed items including strings, none, etc.
    assert evaluate_expression("value in ['abc', none, 10]", 10)  # => True
    assert evaluate_expression(
        "value in ['abc', null, 10]", 10
    )  # => True, "null" => None

    # 2) Compare to literal dict
    # e.g. "value == {'test': 1}"
    my_dict = {"test": 1}
    assert evaluate_expression('value == {"test": 1}', my_dict)  # => True
    assert not evaluate_expression("value == {'test': 2}", my_dict)  # => False
    # Nested dict & lists
    big_dict = {"test": 1, "nested": [3, 4, {"key": False}]}
    assert evaluate_expression(
        "value == {'test':1, 'nested': [3,4, {'key':false}]}", big_dict
    )

    # 3) "in" a dict means "left in dict_keys"
    assert evaluate_expression("value in {'test':1, \"nested\":2}", "nested")  # => True
    assert not evaluate_expression("value in {'test':1, 'nested':2}", "none")

    assert evaluate_expression("'test' == 'test'")
    assert not evaluate_expression("'test' in 'tes_t'")

    assert not evaluate_expression("value not in [false, null, [], 0, '', {}]", None)
    assert not evaluate_expression("value not in [false, null, [], 0, '', {}]", False)
    assert not evaluate_expression("value not in [false, null, [], 0, '', {}]", "")
    assert not evaluate_expression("value not in [false, null, [], 0, '', {}]", 0)
    assert not evaluate_expression("value not in [false, null, [], 0, '', {}]", [])
    assert not evaluate_expression("value not in [false, null, [], 0, '', {}]", {})
    assert not evaluate_expression("value", None)
    assert not evaluate_expression("value", False)
    assert not evaluate_expression("value", "")
    assert not evaluate_expression("value", 0)
    assert not evaluate_expression("value", [])
    assert not evaluate_expression("value", {})
    assert evaluate_expression("value not in [false, null, [], 0, '', {}]", 1)
    assert evaluate_expression("value not in [false, null, [], 0, '', {}]", "test")
    assert evaluate_expression("value not in [false, null, [], 0, '', {}]", True)
    assert evaluate_expression("value not in [false, null, [], 0, '', {}]", [0])
    assert evaluate_expression("value not in [false, null, [], 0, '', {}]", {"test": 1})

    print("All tests passed!")



from enum import Enum
from typing import Annotated, Any, Optional

import httpx
from pydantic import BaseModel
from smartspace.core import Block, Config, Metadata, metadata, step
from smartspace.enums import BlockCategory


class HTTPMethod(str, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"


class ResponseObject(BaseModel):
    content: bytes
    headers: dict[Any, Any]
    body: Any
    status_code: int
    text: str


@metadata(
    description="Performs HTTP requests such as GET, POST, PUT, DELETE, and more.",
    category=BlockCategory.FUNCTION,
    icon="fa-cloud-download-alt",
    label="HTTP request, web API call, REST client, API request, web service call",
)
class HTTPRequest_2_0_0(Block):
    timeout: Annotated[int, Config()] = 30  # Timeout in seconds

    method: Annotated[HTTPMethod, Config()] = HTTPMethod.GET
    headers: Annotated[
        dict[str, Any], Config()
    ] = {}  # NOTE: mutable default is fine if your framework handles it

    @step(output_name="response")
    async def make_request(
        self,
        url: Annotated[str, Metadata(description="The URL to send the request to")],
        body: Annotated[
            Optional[Any],
            Metadata(
                description="Request body for POST, PUT, or PATCH requests. Can be a dict, list, or raw JSON-compatible value."
            ),
        ] = None,
        query_params: Annotated[
            Optional[dict[str, Any]], Metadata(description="Query parameters")
        ] = None,
    ) -> ResponseObject:
        if not url:
            raise ValueError("URL is required")

        query_params = query_params or {}
        json_body = (
            body
            if self.method in {HTTPMethod.POST, HTTPMethod.PUT, HTTPMethod.PATCH}
            else None
        )

        async with httpx.AsyncClient(timeout=self.timeout) as client:
            response = await client.request(
                method=self.method.value,  # ensure str
                url=url,
                headers=self.headers,
                params=query_params,
                json=json_body,
            )

            content_type = response.headers.get("content-type", "").lower()
            response_body = (
                response.json()
                if "application/json" in content_type or "text/json" in content_type
                else response.text
            )

            return ResponseObject(
                status_code=response.status_code,
                headers=dict(response.headers.items()),
                text=response.text,
                content=response.content,
                body=response_body,
            )



from typing import Annotated, Any, Dict, List, Optional

import httpx
from pydantic import BaseModel, Field
from smartspace.core import Block, Config, Metadata, metadata, step
from smartspace.enums import BlockCategory

# Pydantic Models for the Google Custom Search API response


class SearchUrl(BaseModel):
    type: Optional[str] = Field(None, description="MIME type for the response format")
    template: Optional[str] = Field(None, description="URL template for API requests")


class QueryRequest(BaseModel):
    title: Optional[str] = Field(None, description="Title of the query request")
    totalResults: Optional[str] = Field(None, description="Total results returned")
    searchTerms: Optional[str] = Field(None, description="The search query string")
    count: Optional[int] = Field(None, description="Number of results per page")
    startIndex: Optional[int] = Field(
        None, description="The starting index for results"
    )
    inputEncoding: Optional[str] = Field(
        None, description="Input encoding for the query"
    )
    outputEncoding: Optional[str] = Field(
        None, description="Output encoding for the query"
    )
    safe: Optional[str] = Field(None, description="Safe search setting")
    cx: Optional[str] = Field(None, description="Custom search engine ID")


class Queries(BaseModel):
    request: List[QueryRequest] = Field(
        ..., description="Details of the current search query"
    )
    nextPage: Optional[List[QueryRequest]] = Field(
        None, description="Details to retrieve the next page of results"
    )


class SearchInformation(BaseModel):
    searchTime: Optional[float] = Field(
        None, description="Time taken for the search in seconds"
    )
    formattedSearchTime: Optional[str] = Field(
        None, description="Formatted search time"
    )
    totalResults: Optional[str] = Field(
        None, description="Total number of results as a string"
    )
    formattedTotalResults: Optional[str] = Field(
        None, description="Formatted total number of results"
    )


class SearchItem(BaseModel):
    kind: Optional[str] = Field(None, description="Type of the search result item")
    title: Optional[str] = Field(None, description="Title of the search result")
    htmlTitle: Optional[str] = Field(None, description="HTML formatted title")
    link: Optional[str] = Field(None, description="Direct URL to the search result")
    displayLink: Optional[str] = Field(None, description="Simplified display URL")
    snippet: Optional[str] = Field(None, description="Short snippet of the result")
    htmlSnippet: Optional[str] = Field(None, description="HTML formatted snippet")
    formattedUrl: Optional[str] = Field(None, description="User-friendly URL")
    htmlFormattedUrl: Optional[str] = Field(None, description="HTML formatted URL")
    pagemap: Optional[Dict[str, Any]] = Field(
        None, description="Additional structured data if available"
    )


class GoogleSearchResponse(BaseModel):
    kind: Optional[str] = Field(None, description="Response type")
    url: Optional[SearchUrl] = Field(
        None, description="URL details for the API response"
    )
    queries: Optional[Queries] = Field(None, description="Query metadata")
    context: Optional[Dict[str, Any]] = Field(
        None, description="Additional context about the search"
    )
    searchInformation: Optional[SearchInformation] = Field(
        None, description="Performance information for the search"
    )
    items: Optional[List[SearchItem]] = Field(
        None, description="List of search result items"
    )


# Block implementation that uses the Pydantic models
@metadata(
    category=BlockCategory.FUNCTION,
    description="""
    A block that connects to the Google Custom Search API, parses the JSON response 
    into a Pydantic model, and returns the structured result.
    It uses a configuration value 'page' (default=1) to determine which page of results to fetch.
    """,
    icon="fa-search",
)
class GoogleSearch(Block):
    google_api_key: Annotated[str, Config()]
    search_engine_id: Annotated[str, Config()]
    # New config value for the page number (defaults to 1)
    page: Annotated[int, Config()] = 1

    @step(output_name="search_results")
    async def run_search(
        self,
        query: Annotated[
            str,
            Metadata(description="The query string to search on Google"),
        ],
    ) -> GoogleSearchResponse:
        """
        Connects to the Google Custom Search API using the provided API key, search engine ID,
        and the configured page number (defaulting to 1). The method calculates the 'start'
        parameter based on the page number and returns the structured result as a Pydantic model.
        """
        url = "https://www.googleapis.com/customsearch/v1"
        # Calculate the start index: page 1 -> start=1, page 2 -> start=11, etc.
        start_index = (self.page - 1) * 10 + 1
        params = {
            "key": self.google_api_key,
            "cx": self.search_engine_id,
            "q": query,
            "start": start_index,
        }

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(url, params=params)
            response.raise_for_status()
            raw_json = response.json()
            # Use model_validate to convert the raw JSON into our Pydantic model
            parsed_response = GoogleSearchResponse.model_validate(raw_json)
            return parsed_response
        except httpx.HTTPError as e:
            raise ValueError(f"Error while connecting to Google Search API: {e}")



import json
import re
from enum import Enum
from typing import Annotated, Any, List, Union

from jsonpath_ng import JSONPath
from jsonpath_ng.ext import parse
from pydantic import BaseModel

from smartspace.core import (
    Block,
    Config,
    Metadata,
    OperatorBlock,
    Output,
    State,
    metadata,
    step,
)
from smartspace.enums import BlockCategory


@metadata(
    description="This block takes a JSON string or a list of JSON strings and parses them",
    category=BlockCategory.FUNCTION,
    icon="fa-code",
    label="parse JSON, convert JSON string, decode JSON, JSON deserialize, extract JSON data",
)
class ParseJson(OperatorBlock):
    @step(output_name="json")
    async def parse_json(
        self,
        json_string: Annotated[
            Union[str, List[str]],
            Metadata(description="JSON string or list of JSON strings"),
        ],
    ) -> dict[str, Any] | list[dict[str, Any]]:
        if isinstance(json_string, dict):
            return json_string
        elif isinstance(json_string, list):
            results: list[Any] = [json.loads(item) for item in json_string]
            return results
        else:
            result = json.loads(json_string)
            return result


@metadata(
    description="This block removes a specified key from a JSON object. If 'recursive' is set to true, it recursively removes the key from all nested dictionaries; otherwise, it only removes the key from the top level. Returns a copy of the object with the key removed.",
    category=BlockCategory.FUNCTION,
    icon="fa-code",
    label="Remove Key from any json object",
)
class RemoveProperty(OperatorBlock):
    key: Annotated[str, Config()]
    recursive: Annotated[bool, Config()] = False

    @step(output_name="json")
    async def process_object(
        self,
        object: Annotated[
            dict, Metadata(description="Input JSON object as a dictionary")
        ],
    ) -> dict:
        if self.recursive:

            def recursive_remove(item: dict):
                if self.key in item:
                    item.pop(self.key)
                for key, value in list(item.items()):
                    if isinstance(value, dict):
                        recursive_remove(value)

            recursive_remove(object)
        else:
            if self.key in object:
                object.pop(self.key)
        return object


@metadata(
    description="This block retrieves the keys of a JSON object (dictionary). Returns a list of keys.",
    category=BlockCategory.FUNCTION,
    icon="fa-code",
    label="Get Keys from json object",
)
class GetKeys(OperatorBlock):
    @step(output_name="keys")
    async def process_json(
        self,
        object: Annotated[
            dict, Metadata(description="Input JSON object as a dictionary")
        ],
    ) -> List[str]:
        return list(object.keys())


@metadata(
    category=BlockCategory.FUNCTION,
    description="Uses JSONPath to extract data from a JSON object or list. This block will be moved to connection configuration in a future version",
    obsolete=True,
    label="extract JSON field, get JSON value, access JSON data, retrieve JSON property, JSON path query",
    use_instead="Get",
)
class GetJsonField(Block):
    json_field_structure: Annotated[str, Config()]

    @step(output_name="field")
    async def get(self, json_object: Any) -> Any:
        if isinstance(json_object, BaseModel):
            json_object = json.loads(json_object.model_dump_json())
        elif isinstance(json_object, list) and all(
            isinstance(item, BaseModel) for item in json_object
        ):
            json_object = [json.loads(item.model_dump_json()) for item in json_object]

        jsonpath_expr: JSONPath = parse(self.json_field_structure)
        results: List[Any] = [match.value for match in jsonpath_expr.find(json_object)]
        return results


@metadata(
    category=BlockCategory.FUNCTION,
    description="Uses JSONPath to extract data from a JSON object or list.\nJSONPath implementation is from https://pypi.org/project/jsonpath-ng/.",
    icon="fa-search",
    label="get JSON path, query JSON data, extract JSON values, JSON lookup, search JSON",
)
class Get(OperatorBlock):
    path: Annotated[str, Config()]

    @step(output_name="result")
    async def get(self, data: list[Any] | dict[str, Any]) -> Any:
        jsonpath_expr: JSONPath = parse(self.path)
        if isinstance(data, list):
            return [match.value for match in jsonpath_expr.find(data)]
        else:
            results = [match.value for match in jsonpath_expr.find(data)]
            return None if not len(results) else results[0]


class JoinType(Enum):
    INNER = "inner"
    OUTER = "outer"
    LEFT_INNER = "left_inner"
    LEFT_OUTER = "left_outer"
    RIGHT_INNER = "right_inner"
    RIGHT_OUTER = "right_outer"


@metadata(
    category=BlockCategory.FUNCTION,
    description="""
The `Join` block performs advanced join operations between two lists of dictionaries based on a specified key. It merges the data according to the selected join type, similar to SQL join operations, allowing for flexible data integration and transformation.

**Key Features**:

- **Flexible Join Types**: Supports multiple join types, including `INNER`, `LEFT_INNER`, `LEFT_OUTER`, `RIGHT_INNER`, `RIGHT_OUTER`, and `OUTER`.
- **Customizable Key**: Allows specification of the join key.
- **Data Merging**: Combines fields from both left and right records where applicable.
- **Error Handling**: Ensures all records contain the specified key.

**Supported Join Types**:

- **INNER**: Records where the key exists in both left and right lists.
- **LEFT_INNER**: Left records with matching keys in the right list.
- **LEFT_OUTER**: All left records, merging with right records where keys match.
- **RIGHT_INNER**: Right records with matching keys in the left list.
- **RIGHT_OUTER**: All right records, merging with left records where keys match.
- **OUTER**: All records from both lists, merging where keys match.

**Use Cases**:

- Merging datasets from different sources.
- Performing SQL-like join operations in Python.
""",
    icon="fa-link",
    label="join JSON data, SQL-like join, merge datasets, combine JSON lists, data integration",
)
class Join(Block):
    key: Annotated[str, Config()]
    joinType: Annotated[JoinType, Config()] = JoinType.INNER

    @step(output_name="result")
    async def Join(
        self,
        left: list[dict[str, Any]],
        right: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        # Create dictionaries mapping key values to records
        left_dict = {}
        for item in left:
            key_value = item.get(self.key)
            if key_value is None:
                raise KeyError(
                    f"Left item {item} does not contain the key '{self.key}'"
                )
            left_dict[key_value] = item

        right_dict = {}
        for item in right:
            key_value = item.get(self.key)
            if key_value is None:
                raise KeyError(
                    f"Right item {item} does not contain the key '{self.key}'"
                )
            right_dict[key_value] = item

        # Determine the keys to include based on the join type
        if self.joinType == JoinType.INNER:
            keys = set(left_dict.keys()) & set(right_dict.keys())
        elif self.joinType == JoinType.LEFT_INNER:
            keys = set(k for k in left_dict.keys() if k in right_dict)
        elif self.joinType == JoinType.LEFT_OUTER:
            keys = set(left_dict.keys())
        elif self.joinType == JoinType.RIGHT_INNER:
            keys = set(k for k in right_dict.keys() if k in left_dict)
        elif self.joinType == JoinType.RIGHT_OUTER:
            keys = set(right_dict.keys())
        elif self.joinType == JoinType.OUTER:
            keys = set(left_dict.keys()) | set(right_dict.keys())
        else:
            raise ValueError(
                f"Invalid joinType '{self.joinType}'. Must be one of JoinType."
            )

        # Merge the records based on the keys
        result = []
        for key in keys:
            merged_item = {}
            left_item = left_dict.get(key)
            right_item = right_dict.get(key)

            if self.joinType in (
                JoinType.LEFT_OUTER,
                JoinType.LEFT_INNER,
                JoinType.INNER,
                JoinType.OUTER,
            ):
                if left_item:
                    merged_item.update(left_item)
            if self.joinType in (
                JoinType.RIGHT_OUTER,
                JoinType.RIGHT_INNER,
                JoinType.INNER,
                JoinType.OUTER,
            ):
                if right_item:
                    merged_item.update(right_item)

            # Only include items that have data
            if merged_item:
                result.append(merged_item)

        return result


@metadata(
    description="Merges multiple dictionaries into a single object. Accepts only dicts and combines all key-value pairs into one dictionary.",
    category=BlockCategory.MISC,
    icon="fa-cube",
    label="merge objects, combine dictionaries, build object, aggregate key-value pairs",
)
class MergeObjects(Block):
    @step(output_name="object")
    async def build(self, *objects: dict[str, Any]) -> dict[str, Any]:
        merged_object = {}
        for obj in objects:
            merged_object.update(obj)
        return merged_object


@metadata(
    description="Takes in inputs and creates an object containing the inputs",
    category=BlockCategory.MISC,
    icon="fa-cube",
    label="create object, build dictionary, construct object, make key-value map, generate object",
)
class CreateObject(Block):
    @step(output_name="object")
    async def build(self, **properties: Any) -> dict[str, Any]:
        return properties


@metadata(
    category=BlockCategory.FUNCTION,
    label="object builder, json merge, dictionary update, data aggregation, object construction",
    description="Merges objects using dictionary unpacking (similar to jq's merge). Each new object is merged with the existing accumulated object.",
)
class BuildObject(Block):
    """
    A block that merges objects using dictionary unpacking (similar to jq's merge).
    Each new object is merged with the existing accumulated object.
    """

    merged_object: Annotated[dict[str, Any], State()] = {}

    @step(output_name="merged_object")
    async def merge_object(self, obj: dict[str, Any]) -> dict[str, Any]:
        # If a string is passed, try to parse it as JSON
        if isinstance(obj, str):
            cleaned_str = obj.strip()

            # Remove markdown code block syntax if present
            markdown_pattern = r"```(?:json)?\s*([\s\S]*?)\s*```"
            markdown_match = re.search(markdown_pattern, cleaned_str)
            if markdown_match:
                cleaned_str = markdown_match.group(1).strip()

            # Remove potential control characters
            cleaned_str = re.sub(r"[\x00-\x1F\x7F]", "", cleaned_str)

            # Ensure proper JSON formatting for common issues
            if not cleaned_str.startswith("{") and not cleaned_str.startswith("["):
                # Try to find the first occurrence of '{' or '['
                json_start = min(
                    (
                        cleaned_str.find("{")
                        if cleaned_str.find("{") >= 0
                        else float("inf")
                    ),
                    (
                        cleaned_str.find("[")
                        if cleaned_str.find("[") >= 0
                        else float("inf")
                    ),
                )
                if json_start != float("inf"):
                    cleaned_str = cleaned_str[json_start:]

            obj = json.loads(cleaned_str)

        self.merged_object = {**self.merged_object, **obj}
        return self.merged_object


@metadata(
    description="Takes in an object and sends each key-value pair to the corresponding output",
    category=BlockCategory.MISC,
    icon="fa-th-large",
    label="unpack object, extract object properties, decompose dictionary, spread object, distribute fields",
)
class UnpackObject(Block):
    properties: dict[str, Output[dict[str, Any]]]

    @step()
    async def unpack(self, object: dict[str, Any]):
        for name, value in object.items():
            if name in self.properties:
                self.properties[name].send(value)



from typing import Annotated, Any, Generic, TypeVar

from smartspace.core import Block, Output, State, metadata, step

ValueT = TypeVar("ValueT")


@metadata(
    description="Buffers values in a list and outputs them one at a time.",
    icon="fa-database",
    label="buffer, queue, buffer block, queue block, buffer list, queue list",
    obsolete=True,
    deprecated_reason="This block will be deprecated in a future version. ",
)
class Buffer(Block, Generic[ValueT]):
    values: Annotated[list[ValueT], State()] = []
    ready: Annotated[bool, State()] = True
    output: Output[ValueT]

    @step()
    async def value(self, value: ValueT):
        self.values.append(value)
        await self._inner()

    @step()
    async def next(self, next: Any):
        self.ready = True
        await self._inner()

    async def _inner(self):
        if not self.ready:
            return

        if len(self.values):
            self.ready = False
            v = self.values.pop(0)
            self.output.send(v)



import inspect
from typing import Annotated, Callable

from pydantic import TypeAdapter
from typing_extensions import get_origin


def _issubclass(cls, base):
    return inspect.isclass(cls) and issubclass(get_origin(cls) or cls, base)


def _get_type_adapter(annotation: type) -> TypeAdapter:
    if get_origin(annotation) is Annotated:
        return TypeAdapter(annotation.__args__[0])
    elif annotation is inspect.Parameter.empty:
        return TypeAdapter(object)
    else:
        return TypeAdapter(annotation)


def get_return_type(callable: Callable):
    signature = inspect.signature(callable)
    return (
        signature.return_annotation
        if signature.return_annotation != signature.empty
        else None
    )


def get_parameter_names_and_types(callable: Callable):
    signature = inspect.signature(callable)
    return [
        (name, param.annotation)
        for name, param in signature.parameters.items()
        if name != "self"
    ]



import re
from typing import Annotated

from smartspace.core import Block, Config, metadata, step
from smartspace.enums import BlockCategory


@metadata(
    category=BlockCategory.FUNCTION,
    description="regex-based pattern matching on a string input, returning a list of all matches found.",
    icon="fa-text-width",
    label="regex match, pattern matching, regular expression, string pattern, text extraction",
)
class RegexMatch(Block):
    """
    Input:
        1. a string input
        2. a regex expression
    Output: a list of match results of a regex expression
    """

    regex: Annotated[str, Config()] = (
        r".*"  # Default pattern to match the entire string
    )

    @step(output_name="result")
    async def regex_match(self, input_strings: str) -> list[str]:
        try:
            pattern = re.compile(self.regex)
            match = pattern.findall(input_strings)
            if len(match) == 0:
                return ["No match found"]
            return match
        except Exception as e:
            error_message = f"Error: {e}"
            return [error_message]



import pytest

from smartspace.blocks.regex_match import RegexMatch


@pytest.mark.asyncio
async def test_regex_match_default_pattern():
    regex_block = RegexMatch()
    input_string = "Hello, World!"
    result = await regex_block.regex_match(input_string)
    assert result == ["Hello, World!", ""]


@pytest.mark.asyncio
async def test_regex_match_custom_pattern():
    regex_block = RegexMatch()
    regex_block.regex = r"\b\w+\b"  # Match words
    input_string = "Hello, World! How are you?"
    result = await regex_block.regex_match(input_string)
    assert result == ["Hello", "World", "How", "are", "you"]


@pytest.mark.asyncio
async def test_regex_match_no_match():
    regex_block = RegexMatch()
    regex_block.regex = r"\d+"  # Match numbers
    input_string = "No numbers here"
    result = await regex_block.regex_match(input_string)
    assert result == ["No match found"]


@pytest.mark.asyncio
async def test_regex_match_invalid_pattern():
    regex_block = RegexMatch()
    regex_block.regex = r"["  # Invalid regex pattern
    input_string = "Test string"
    result = await regex_block.regex_match(input_string)
    assert result[0].startswith("Error: ")


@pytest.mark.asyncio
async def test_regex_match_empty_input():
    regex_block = RegexMatch()
    input_string = ""
    result = await regex_block.regex_match(input_string)
    assert result == [""]


@pytest.mark.asyncio
async def test_regex_match_multiple_matches():
    regex_block = RegexMatch()
    regex_block.regex = r"\b\w{3}\b"  # Match 3-letter words
    input_string = "The cat and dog are pets"
    result = await regex_block.regex_match(input_string)
    assert result == ["The", "cat", "and", "dog", "are"]



from enum import Enum


class BlockCategory(Enum):
    AGENT = {"name": "Agent", "description": "An entity that performs actions"}
    FUNCTION = {"name": "Function", "description": "A callable entity"}
    DATA = {"name": "Data", "description": "A data entity"}
    CUSTOM = {"name": "Custom", "description": "A custom entity"}
    MISC = {"name": "Misc", "description": "Doesnt belong to any category"}


class FlowVariableAccess(Enum):
    NONE = "None"
    READ = "Read"
    WRITE = "Write"


class BlockClass(Enum):
    MODEL = "Model"
    OPERATOR = "Operator"


class ChannelEvent(Enum):
    DATA = "Data"
    CLOSE = "Close"


class ChannelState(Enum):
    OPEN = "Open"
    CLOSED = "Closed"


class BlockScope(Enum):
    WORKSPACE = "WorkSpace"
    DATASET = "DataSet"


class InputDisplayType(Enum):
    TEMPLATEOBJECT = "TemplateObject"



import json
from typing import Annotated, Any, Generic, TypeVar, cast

from pydantic import BaseModel, ConfigDict

from smartspace.core import Config, GenericSchema, OperatorBlock, metadata, step
from smartspace.enums import BlockCategory

ItemT = TypeVar("ItemT")


@metadata(
    description="Takes in any input and will attempt to convert the input to the specified schema. If the convert config is unticked, it will not attempt to convert the value and will instead just output the input.",
    category=BlockCategory.MISC,
    icon="fa-sync-alt",
    label="cast type, convert data, transform format, change type, typecast value",
)
class Cast(OperatorBlock, Generic[ItemT]):
    schema: GenericSchema[ItemT]
    convert: Annotated[bool, Config()] = True

    @step(output_name="result")
    async def cast(self, item: Any) -> ItemT:
        if not self.convert:
            return item
        if "type" not in self.schema:
            return item

        return self._cast(item, self.schema)

    def _cast(self, item: Any, schema: dict[str, Any]) -> Any:
        if "type" not in schema:
            return item

        if schema["type"] == "array":
            return cast(ItemT, [self._cast(i, schema["items"]) for i in item])

        if schema["type"] == "object":
            if isinstance(item, str):
                item = json.loads(item)

            if len(schema) == 1:
                return item

            class M(BaseModel):
                model_config = ConfigDict(
                    json_schema_extra=schema,
                )

            if isinstance(item, dict):
                return M.model_validate(item)
            elif isinstance(item, str):
                return M.model_validate_json(item)
            else:
                raise ValueError(f"Cannot cast type '{type(item)}' to object")

        elif schema["type"] == "string":
            if isinstance(item, str):
                return item
            else:
                return json.dumps(item, indent=2)

        elif schema["type"] == "number":
            if isinstance(item, (int, float)):
                return item
            else:
                try:
                    return float(item)
                except (TypeError, ValueError) as e:
                    raise ValueError(f"Cannot convert '{item}' to float.") from e
        else:
            return item



import abc
import asyncio
import asyncio.queues
import contextvars
import copy
import enum
import inspect
import json
import types
import typing
from typing import (
    Annotated,
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Concatenate,
    Generic,
    Literal,
    Mapping,
    NamedTuple,
    ParamSpec,
    TypeVar,
    cast,
)

import semantic_version
from more_itertools import first
from pydantic import BaseModel, ConfigDict, TypeAdapter, ValidationError
from pydantic._internal._generics import get_args, get_origin

from smartspace.enums import (
    BlockCategory,
    BlockClass,
    BlockScope,
    ChannelEvent,
    InputDisplayType,
)
from smartspace.models import (
    BlockErrorModel,
    BlockInterface,
    BlockPinRef,
    BlockRunMessage,
    FlowContext,
    InputChannel,
    InputPinInterface,
    InputValue,
    OutputChannelMessage,
    OutputPinInterface,
    OutputValue,
    PinRedirect,
    PinType,
    PortInterface,
    PortType,
    SmartSpaceWorkspace,
    StateInterface,
    StateValue,
    ThreadMessage,
)
from smartspace.utils.utils import _get_type_adapter, _issubclass

B = TypeVar("B", bound="Block")
S = TypeVar("S")
T = TypeVar("T")
R = TypeVar("R")
P = ParamSpec("P")


def _get_pin_type_from_parameter_kind(kind: inspect._ParameterKind) -> PinType:
    if (
        kind == inspect._ParameterKind.KEYWORD_ONLY
        or kind == inspect._ParameterKind.POSITIONAL_OR_KEYWORD
    ):
        return PinType.SINGLE
    elif kind == inspect._ParameterKind.VAR_POSITIONAL:
        return PinType.LIST
    elif kind == inspect._ParameterKind.VAR_KEYWORD:
        return PinType.DICTIONARY
    else:
        raise Exception(f"Invalid parameter kind {kind}")


class FunctionPins(NamedTuple):
    inputs: dict[str, InputPinInterface]
    input_adapters: dict[str, TypeAdapter]
    output: tuple[OutputPinInterface | None, TypeAdapter | None]
    generics: dict[str, TypeAdapter]


class InputInfo(NamedTuple):
    type: type
    is_channel: bool


def check_type_is_input_channel(annotation: type) -> InputInfo:
    if get_origin(annotation) is InputChannel:
        args = get_args(annotation)
        if not args or len(args) != 1:
            raise Exception("Input channels must have exactly one type.")

        return InputInfo(type=args[0], is_channel=True)
    else:
        return InputInfo(type=annotation, is_channel=False)


def _get_generics(t: type) -> list[TypeVar]:
    bases = getattr(t, "__orig_bases__", None)
    if bases is None or len(bases) == 0:
        return []

    for base in bases:
        if get_origin(base) == Generic:
            return list(get_args(base))

    return []


def _get_function_pins(fn: Callable, port_name: str | None = None) -> FunctionPins:
    signature = inspect.signature(fn)
    inputs: dict[str, InputPinInterface] = {}
    input_adapters: dict[str, TypeAdapter] = {}
    generics: dict[str, TypeAdapter] = {}

    for name, param in signature.parameters.items():
        if name == "self":
            continue

        annotations = getattr(param.annotation, "__metadata__", [])
        metadata = first(
            (metadata.data for metadata in annotations if type(metadata) is Metadata),
            {},
        )

        if param.default == inspect._empty:
            default = None
            required = True
        else:
            default = param.default
            required = False

        input_type, is_channel = check_type_is_input_channel(param.annotation)

        type_adapter, schema, _generics = _get_json_schema_with_generics(input_type)
        generics.update(_generics)

        inputs[name] = InputPinInterface(
            metadata=metadata,
            json_schema=schema,
            sticky=any(
                [metadata.sticky for metadata in annotations if type(metadata) is Input]
            ),
            type=_get_pin_type_from_parameter_kind(param.kind),
            default=default,
            required=required,
            generics={
                name: BlockPinRef(
                    port=port_name if port_name else name,
                    pin=name if port_name else "",
                )
                for name in _generics.keys()
            },
            channel=is_channel,
            virtual=False,
        )
        input_adapters[name] = type_adapter

    if signature.return_annotation != signature.empty:
        annotations = getattr(signature.return_annotation, "__metadata__", [])
        metadata = first(
            (metadata.data for metadata in annotations if type(metadata) is Metadata),
            {},
        )

        if get_origin(signature.return_annotation) == OutputChannel:
            args = get_args(signature.return_annotation)
            if not args or len(args) != 1:
                raise Exception("Outputs must have exactly one type.")

            output_type: type = args[0]
            is_channel = True
        else:
            output_type = signature.return_annotation
            is_channel = False

        output_type_adapter, schema, _generics = _get_json_schema_with_generics(
            output_type
        )
        generics.update(_generics)

        output = (
            OutputPinInterface(
                metadata=metadata,
                json_schema=schema,
                type=PinType.SINGLE,
                generics={
                    name: BlockPinRef(
                        port=port_name if port_name else name,
                        pin=name if port_name else "",
                    )
                    for name in _generics.keys()
                },
                channel=is_channel,
                channel_group_id=port_name if is_channel else None,
            ),
            output_type_adapter,
        )
    else:
        output = (None, None)

    return FunctionPins(
        output=output,
        inputs=inputs,
        input_adapters=input_adapters,
        generics=generics,
    )


class ToolPins(NamedTuple):
    input: tuple[InputPinInterface | None, TypeAdapter | None]
    outputs: dict[str, OutputPinInterface]
    output_adapters: dict[str, TypeAdapter]
    generics: dict[str, TypeAdapter]


def _get_tool_pins(
    fn: Callable,
    port_name: str | None = None,
    generic_names: list[str] | None = None,
) -> ToolPins:
    generic_names = generic_names or []
    signature = inspect.signature(fn)
    outputs: dict[str, OutputPinInterface] = {}
    output_adapters: dict[str, TypeAdapter] = {}
    generics: dict[str, TypeAdapter] = {}

    for name, param in signature.parameters.items():
        if name == "self":
            continue

        annotations = getattr(param.annotation, "__metadata__", [])
        metadata = first(
            (metadata.data for metadata in annotations if type(metadata) is Metadata),
            {},
        )

        output_type = param.annotation

        type_adapter, schema, _generics = _get_json_schema_with_generics(output_type)
        pin_generics: dict[str, BlockPinRef] = {}

        for generic_name in _generics.keys():
            if generic_name in generic_names:
                pin_generics[generic_name] = BlockPinRef(
                    port=port_name if port_name else name,
                    pin=generic_name if port_name else "",
                )
                generics[generic_name] = _generics[generic_name]
            else:
                pin_generics[generic_name] = BlockPinRef(
                    port=generic_name,
                    pin="",
                )

        outputs[name] = OutputPinInterface(
            metadata=metadata,
            json_schema=schema,
            type=_get_pin_type_from_parameter_kind(param.kind),
            generics=pin_generics,
            channel=True,
            channel_group_id=port_name,  # Tool outputs should always be channels so multiple tool calls in a function execution have different scopes
        )

        output_adapters[name] = type_adapter

    if signature.return_annotation != signature.empty:
        annotations = getattr(signature.return_annotation, "__metadata__", [])
        metadata = first(
            (metadata.data for metadata in annotations if type(metadata) is Metadata),
            {},
        )
        input_type, is_channel = check_type_is_input_channel(
            signature.return_annotation
        )

        input_type_adapter, schema, _generics = _get_json_schema_with_generics(
            input_type
        )

        pin_generics: dict[str, BlockPinRef] = {}
        for generic_name in _generics.keys():
            if generic_name in generic_names:
                pin_generics[generic_name] = BlockPinRef(
                    port=port_name if port_name else name,
                    pin=generic_name if port_name else "",
                )
                generics[generic_name] = _generics[generic_name]
            else:
                pin_generics[generic_name] = BlockPinRef(
                    port=generic_name,
                    pin="",
                )

        _input = (
            InputPinInterface(
                metadata=metadata,
                json_schema=schema,
                type=PinType.SINGLE,
                sticky=False,
                required=False,
                default=None,
                generics=pin_generics,
                channel=is_channel,
                virtual=True,
            ),
            input_type_adapter,
        )
    else:
        _input = (None, None)

    return ToolPins(
        outputs=outputs,
        output_adapters=output_adapters,
        input=_input,
        generics=generics,
    )


def _get_default(cls, field_name) -> tuple[bool, Any]:
    no_default = "__no_default__"
    default_value = getattr(cls, field_name, no_default)
    if default_value is not no_default:
        return (True, default_value)

    return (False, None)


class Metadata:
    def __init__(
        self,
        **kwargs,
    ):
        self.data = kwargs


GenericSchemaT = TypeVar("GenericSchemaT")


class GenericSchema(Generic[GenericSchemaT], dict[str, Any]): ...


class Config: ...


def _get_all_bases(cls: type):
    bases: list[type] = []

    cls_bases = set(
        getattr(cls, "__orig_bases__", tuple())
        + getattr(cls, "__bases__", tuple())
        + getattr(get_origin(cls), "__bases__", tuple())
    )

    for base in cls_bases:
        if base in bases:
            continue

        bases.append(base)
        bases.extend(_get_all_bases(base))

    return bases


def _get_input_pin_from_metadata(
    field_type: type,
    pin_type: PinType,
    port_name: str,
    field_name: str,
    parent: type | None = None,
) -> tuple[tuple[InputPinInterface | None, TypeAdapter | None], dict[str, TypeAdapter]]:
    config: Config | None = None
    _input: Input | None = None
    state: State | None = None
    metadata: dict[str, Any] = {}

    is_input_channel = get_origin(field_type) is InputChannel
    args = get_args(field_type)
    if is_input_channel:
        if not args or len(args) != 1:
            raise Exception("Input channels must have exactly one type.")

        input_type: type = args[0]
    else:
        input_type = field_type if not args or len(args) == 0 else args[0]
        for m in getattr(field_type, "__metadata__", []):
            if isinstance(m, Config):
                config = m

            if isinstance(m, Input):
                _input = m

            if isinstance(m, State):
                state = m

            if isinstance(m, Metadata):
                metadata = m.data

        matches = len([True for i in [config, _input, state] if i is not None])

        if matches > 1:
            raise ValueError(
                "Fields can only be annotated with one of Config(), Input(), and State()"
            )

        if matches == 0:
            return (None, None), {}

        if state:
            return (None, None), {}

        if config and "config" not in metadata:
            metadata["config"] = True

    default = None
    required = True
    if pin_type == PinType.SINGLE:
        if parent is None:
            raise ValueError(
                f"'parent' must be given when getting the interface for a {pin_type} pin"
            )

        has_default, default_value = _get_default(parent, field_name)
        if config and not has_default and input_type is bool:
            default = False
            required = False

        if has_default:
            required = False
            default = default_value

    type_adapter, schema, _generics = _get_json_schema_with_generics(input_type)

    return (
        (
            InputPinInterface(
                metadata=metadata,
                sticky=config is not None or (_input and _input.sticky) or False,
                json_schema=schema,
                type=pin_type,
                generics={
                    name: BlockPinRef(
                        port=name if port_name == field_name else port_name,
                        pin="" if port_name == field_name else name,
                    )
                    for name in _generics.keys()
                },
                default=default,
                required=required,
                channel=is_input_channel,
                virtual=False,
            ),
            type_adapter,
        ),
        _generics,
    )


def _get_state_from_metadata(
    field_type: type,
    field_name: str,
    block_type: "type[Block]",
) -> StateInterface | None:
    state: State | None = None
    metadata: dict[str, Any] = {}

    for m in getattr(field_type, "__metadata__", []):
        if isinstance(m, State):
            state = m

        if isinstance(m, Metadata):
            metadata = m.data

    if state is None:
        return None

    has_default, default = _get_default(block_type, field_name)
    if not has_default:
        raise ValueError("State() attributes must have a default value")

    state_type = get_args(field_type)[0]
    block_type._state_type_adapters[field_name] = _get_type_adapter(state_type)

    return StateInterface(
        metadata=metadata,
        scope=[
            BlockPinRef(port=state.step_id, pin=p)
            for p in state.input_ids or []
            if state.step_id
        ],
        default=default,
    )


def _map_type_vars(
    original_type: type,
    mode: Literal["schema", "validation"],
) -> tuple[type, dict[TypeVar, TypeAdapter]]:
    type_var_defs: dict[TypeVar, TypeAdapter] = {}

    def _inner(new_type: type | TypeVar, depth: int) -> type:
        origin = get_origin(new_type)
        if origin == Annotated:
            args = get_args(new_type)
            if not args:
                return cast(type, new_type)

            new_type = cast(type, args[0])
            return _inner(new_type, depth + 1)

        if isinstance(new_type, TypeVar):

            class TempTypeVarModel2(BaseModel):
                model_config = ConfigDict(title=new_type.__name__)
                if mode == "validation":
                    __pydantic_core_schema__ = {"type": "any"}

            type_var_defs[new_type] = TypeAdapter(new_type)
            return TempTypeVarModel2

        if depth > 10:
            return new_type

        new_args = []
        args = get_args(new_type)
        if args:
            for arg in args:
                if isinstance(arg, TypeVar):
                    if arg not in type_var_defs:

                        class TempTypeVarModel(BaseModel):
                            model_config = ConfigDict(title=arg.__name__)
                            if mode == "validation":
                                __pydantic_core_schema__ = {"type": "any"}

                        type_var_defs[arg] = TypeAdapter(arg)

                    new_args.append(TempTypeVarModel)
                else:
                    new_args.append(_inner(arg, depth + 1))

            if origin:
                if origin is types.UnionType:
                    origin = typing.Union

                class_getitem = getattr(origin, "__class_getitem__", None)
                if class_getitem:
                    n = class_getitem(tuple(new_args))
                    return n

                getitem = getattr(origin, "__getitem__", None)
                if getitem:
                    n = getitem(tuple(new_args))
                    return n

                return new_type
            return new_type

        return new_type

    type_with_generics_replaced = _inner(original_type, 0)
    return type_with_generics_replaced, type_var_defs


class JsonSchemaWithGenerics(NamedTuple):
    type_adapter: TypeAdapter
    schema: dict[str, Any]
    generics: dict[str, TypeAdapter]


def _get_json_schema_with_generics(t: type) -> JsonSchemaWithGenerics:
    new_t, type_var_map = _map_type_vars(t, mode="schema")
    generics = {name.__name__: adapter for name, adapter in type_var_map.items()}
    json_schema = TypeAdapter(Any if new_t == inspect._empty else new_t).json_schema()

    new_t, _ = _map_type_vars(t, mode="validation")
    type_adapter = TypeAdapter(Any if new_t == inspect._empty else new_t)

    if "$defs" in json_schema:
        definitions: dict[str, dict[str, Any]] = json_schema["$defs"]
        new_definitions: dict[str, dict[str, Any]] = {}

        json_schema_str = json.dumps(json_schema)
        for name, definition in definitions.items():
            if "TempTypeVarModel" in name and "title" in definition:
                title = definition["title"]
                new_definitions[title] = {}
                json_schema_str = json_schema_str.replace(name, title)
            else:
                new_definitions[name] = definition

        json_schema = json.loads(json_schema_str)
        json_schema["$defs"] = new_definitions

    elif "title" in json_schema:
        title = json_schema["title"]
        if title in generics:
            json_schema = {"$defs": {title: {}}, "$ref": "#/$defs/" + title}

    return JsonSchemaWithGenerics(
        type_adapter=type_adapter,
        schema=json_schema,
        generics=generics,
    )


class PinsSet(NamedTuple):
    inputs: dict[str, InputPinInterface]
    outputs: dict[str, OutputPinInterface]
    generics: dict[str, TypeAdapter]


def _get_pins(
    cls_annotation: type,
    port_name: str,
    block_type: "type[Block]",
) -> PinsSet:
    cls_metadata = getattr(cls_annotation, "__metadata__", [])
    if len(cls_metadata):
        cls = get_args(cls_annotation)[0]
        metadata = first([a.data for a in cls_metadata if isinstance(a, Metadata)], {})
    else:
        cls = cls_annotation
        metadata = {}

    all_bases = _get_all_bases(cls) + [cls, cls_annotation]

    inputs: dict[str, InputPinInterface] = {}
    outputs: dict[str, OutputPinInterface] = {}
    generics: dict[str, TypeAdapter] = {}

    for base_type in all_bases:
        o = get_origin(base_type)
        if o is Output or o is OutputChannel:
            args = get_args(base_type)
            if not args or len(args) != 1:
                raise Exception("Outputs must have exactly one type.")

            type_adapter, schema, _generics = _get_json_schema_with_generics(args[0])
            generics.update(_generics)

            outputs[""] = OutputPinInterface(
                metadata=metadata,
                json_schema=schema,
                type=PinType.SINGLE,
                generics={
                    name: BlockPinRef(port=name, pin="") for name in _generics.keys()
                },
                channel=o is OutputChannel,
                channel_group_id=f"{port_name}" if o is OutputChannel else None,
            )

    if _issubclass(cls_annotation, Tool):
        tool_type = cast(Tool, cls_annotation)
        generics_on_this_tool = _get_generics(cls_annotation)

        (_input, input_adapter), _outputs, output_adapters, _generics = _get_tool_pins(
            tool_type.run,
            port_name=port_name,
            generic_names=[t.__name__ for t in generics_on_this_tool],
        )

        outputs.update(_outputs)
        for name, adapter in output_adapters.items():
            block_type._set_output_pin_type_adapter(port_name, name, adapter)

        if _input and input_adapter:
            block_type._set_input_pin_type_adapter(port_name, "return", input_adapter)
            inputs["return"] = _input

        for generic_name, generic_schema in _generics.items():
            type_adapter = TypeAdapter(dict[str, Any])
            block_type._set_input_pin_type_adapter(
                port_name, generic_name, type_adapter
            )

            inputs[generic_name] = InputPinInterface(
                metadata={"generic": True, "hidden": True},
                sticky=True,
                json_schema=type_adapter.json_schema(),
                generics={},
                type=PinType.SINGLE,
                required=False,
                default=generic_schema.json_schema(),
                channel=False,
                virtual=False,
            )

    (input_pin, input_adapter), _generics = _get_input_pin_from_metadata(
        base_type,
        port_name=port_name,
        field_name=port_name,
        parent=block_type,
        pin_type=PinType.SINGLE,
    )
    if isinstance(input_pin, InputPinInterface) and input_adapter:
        inputs[""] = input_pin
        block_type._set_input_pin_type_adapter(port_name, "", input_adapter)
        generics.update(_generics)

    annotations = {}
    for base in all_bases:
        annotations.update(getattr(base, "__annotations__", {}))
    annotations.update(**getattr(cls, "__annotations__", {}))

    for field_name, field_annotation in annotations.items():
        field_metadata = getattr(field_annotation, "__metadata__", [])
        if len(field_metadata):
            field_type = get_args(field_annotation)[0]
            metadata = first(
                [a.data for a in field_metadata if isinstance(a, Metadata)], {}
            )
        else:
            metadata = {}
            field_type = field_annotation

        o = get_origin(field_type)
        if o is Output or o is OutputChannel:
            args = get_args(field_type)
            if not args or len(args) != 1:
                raise Exception("Outputs must have exactly one type.")

            type_adapter, schema, _generics = _get_json_schema_with_generics(args[0])
            block_type._set_output_pin_type_adapter(port_name, field_name, type_adapter)
            for generic_name, generic_schema in _generics.items():
                type_adapter = TypeAdapter(dict[str, Any])
                block_type._set_input_pin_type_adapter(
                    port_name, generic_name, type_adapter
                )

                inputs[generic_name] = InputPinInterface(
                    metadata={"generic": True, "hidden": True},
                    sticky=True,
                    json_schema=type_adapter.json_schema(),
                    generics={},
                    type=PinType.SINGLE,
                    required=False,
                    default=generic_schema.json_schema(),
                    channel=False,
                    virtual=False,
                )

            outputs[field_name] = OutputPinInterface(
                metadata=metadata,
                json_schema=schema,
                type=PinType.SINGLE,
                generics={
                    name: BlockPinRef(port=port_name, pin=name)
                    for name in _generics.keys()
                },
                channel=o is OutputChannel,
                channel_group_id=f"{port_name}.{name}" if o is OutputChannel else None,
            )

        elif o is dict:
            dict_args = get_args(field_type)
            if dict_args:
                item_type: type = dict_args[1]

                is_output = _issubclass(item_type, Output)
                is_output_channel = not is_output and _issubclass(
                    item_type, OutputChannel
                )

                if is_output or is_output_channel:
                    if dict_args[0] is not str:
                        raise TypeError("Output dictionaries must have str keys")

                    args = get_args(field_type)
                    if not args or len(args) != 1:
                        raise Exception("Outputs must have exactly one type.")

                    type_adapter, schema, _generics = _get_json_schema_with_generics(
                        args[0]
                    )
                    block_type._set_output_pin_type_adapter(
                        port_name, field_name, type_adapter
                    )
                    for generic_name, generic_schema in _generics.items():
                        type_adapter = TypeAdapter(dict[str, Any])
                        block_type._set_input_pin_type_adapter(
                            port_name, generic_name, type_adapter
                        )

                        inputs[generic_name] = InputPinInterface(
                            metadata={"generic": True, "hidden": True},
                            sticky=True,
                            json_schema=type_adapter.json_schema(),
                            generics={},
                            type=PinType.SINGLE,
                            required=False,
                            default=generic_schema.json_schema(),
                            channel=False,
                            virtual=False,
                        )

                    outputs[field_name] = OutputPinInterface(
                        metadata=metadata,
                        json_schema=schema,
                        type=PinType.DICTIONARY,
                        generics={
                            name: BlockPinRef(port=port_name, pin=name)
                            for name in _generics.keys()
                        },
                        channel=is_output_channel,
                        channel_group_id=f"{port_name}.{name}"
                        if is_output_channel
                        else None,
                    )
                else:
                    (input_pin, input_adapter), _generics = (
                        _get_input_pin_from_metadata(
                            item_type,
                            port_name=port_name,
                            field_name=field_name,
                            parent=cls,
                            pin_type=PinType.DICTIONARY,
                        )
                    )
                    if isinstance(input_pin, InputPinInterface) and input_adapter:
                        block_type._set_input_pin_type_adapter(
                            port_name, field_name, input_adapter
                        )
                        inputs[field_name] = input_pin
                        for generic_name, generic_schema in _generics.items():
                            type_adapter = TypeAdapter(dict[str, Any])
                            block_type._set_input_pin_type_adapter(
                                port_name, generic_name, type_adapter
                            )

                            inputs[generic_name] = InputPinInterface(
                                metadata={"generic": True, "hidden": True},
                                sticky=True,
                                json_schema=type_adapter.json_schema(),
                                generics={},
                                type=PinType.SINGLE,
                                required=False,
                                default=generic_schema.json_schema(),
                                channel=False,
                                virtual=False,
                            )

        elif o is list:
            list_args = get_args(field_type)
            if list_args:
                item_type: type = list_args[0]

                is_output = _issubclass(item_type, Output)
                is_output_channel = not is_output and _issubclass(
                    item_type, OutputChannel
                )

                if is_output or is_output_channel:
                    args = get_args(field_type)
                    if not args or len(args) != 1:
                        raise Exception("Outputs must have exactly one type.")

                    type_adapter, schema, _generics = _get_json_schema_with_generics(
                        args[0]
                    )
                    block_type._set_output_pin_type_adapter(
                        port_name, field_name, type_adapter
                    )

                    for generic_name, generic_schema in _generics.items():
                        type_adapter = TypeAdapter(dict[str, Any])
                        block_type._set_input_pin_type_adapter(
                            port_name, generic_name, type_adapter
                        )

                        inputs[generic_name] = InputPinInterface(
                            metadata={"generic": True, "hidden": True},
                            sticky=True,
                            json_schema=type_adapter.json_schema(),
                            generics={},
                            type=PinType.SINGLE,
                            required=False,
                            default=generic_schema.json_schema(),
                            channel=False,
                            virtual=False,
                        )

                    outputs[field_name] = OutputPinInterface(
                        metadata=metadata,
                        json_schema=schema,
                        type=PinType.LIST,
                        generics={
                            name: BlockPinRef(port=port_name, pin=name)
                            for name in _generics.keys()
                        },
                        channel=is_output_channel,
                        channel_group_id=f"{port_name}.{name}"
                        if is_output_channel
                        else None,
                    )
                else:
                    (input_pin, input_adapter), _generics = (
                        _get_input_pin_from_metadata(
                            item_type,
                            port_name=port_name,
                            field_name=field_name,
                            parent=cls,
                            pin_type=PinType.LIST,
                        )
                    )
                    if isinstance(input_pin, InputPinInterface) and input_adapter:
                        block_type._set_input_pin_type_adapter(
                            port_name, field_name, input_adapter
                        )
                        inputs[field_name] = input_pin
                        for generic_name, generic_schema in _generics.items():
                            type_adapter = TypeAdapter(dict[str, Any])
                            block_type._set_input_pin_type_adapter(
                                port_name, generic_name, type_adapter
                            )

                            inputs[generic_name] = InputPinInterface(
                                metadata={"generic": True, "hidden": True},
                                sticky=True,
                                json_schema=type_adapter.json_schema(),
                                generics={},
                                type=PinType.SINGLE,
                                required=False,
                                default=generic_schema.json_schema(),
                                channel=False,
                                virtual=False,
                            )

        (input_pin, input_adapter), _generics = _get_input_pin_from_metadata(
            field_annotation,
            port_name=port_name,
            field_name=field_name,
            parent=cls,
            pin_type=PinType.SINGLE,
        )
        if isinstance(input_pin, InputPinInterface) and input_adapter:
            inputs[field_name] = input_pin
            block_type._set_input_pin_type_adapter(port_name, field_name, input_adapter)
            for generic_name, generic_schema in _generics.items():
                type_adapter = TypeAdapter(dict[str, Any])
                block_type._set_input_pin_type_adapter(
                    port_name, generic_name, type_adapter
                )

                inputs[generic_name] = InputPinInterface(
                    metadata={"generic": True, "hidden": True},
                    sticky=True,
                    json_schema=type_adapter.json_schema(),
                    generics={},
                    type=PinType.SINGLE,
                    required=False,
                    default=generic_schema.json_schema(),
                    channel=False,
                    virtual=False,
                )

    for field_name, field_annotation in annotations.items():
        field_metadata = getattr(field_annotation, "__metadata__", [])
        if len(field_metadata):
            field_type = get_args(field_annotation)[0]
            metadata = first(
                [a.data for a in field_metadata if isinstance(a, Metadata)], {}
            )
        else:
            metadata = {}
            field_type = field_annotation

        origin = get_origin(field_type)

        if origin is GenericSchema:
            type_var = get_args(field_type)[0]
            generic_name = type_var.__name__
            if generic_name in inputs:
                inputs[field_name] = inputs[generic_name]
                del inputs[generic_name]

            has_default, default = _get_default(block_type, field_name)
            if has_default:
                inputs[field_name].default = default

            inputs[field_name].metadata["hidden"] = False
            inputs[field_name].metadata["config"] = True
            inputs[field_name].metadata.update(metadata)

            block_type._input_pin_type_adapters[port_name][field_name] = (
                block_type._input_pin_type_adapters[port_name][generic_name]
            )
            del block_type._input_pin_type_adapters[port_name][generic_name]

            for pin in list(inputs.values()) + list(outputs.values()):
                for g, pin_ref in pin.generics.items():
                    if (
                        g == generic_name
                        and pin_ref.port == port_name
                        and pin_ref.pin == generic_name
                    ):
                        pin.generics[g] = BlockPinRef(port=port_name, pin=field_name)

    return PinsSet(inputs, outputs, generics)


class PortsAndState(NamedTuple):
    ports: dict[str, PortInterface]
    state: dict[str, StateInterface]


def _get_ports_and_state(block_type: "type[Block]") -> PortsAndState:
    annotations = {}
    for base in _get_all_bases(block_type):
        base_annotations = getattr(base, "__annotations__", {})
        annotations.update(base_annotations)
    annotations.update(**block_type.__annotations__)

    ports: dict[str, PortInterface] = {}
    state: dict[str, StateInterface] = {}
    generics: dict[str, TypeAdapter] = {}

    for port_name, port_annotation in annotations.items():
        port_annotations = getattr(port_annotation, "__metadata__", None)
        if port_annotations:
            field_type = get_args(port_annotation)[0]
            metadata = first(
                [a.data for a in port_annotations if isinstance(a, Metadata)], {}
            )
        else:
            field_type = port_annotation
            metadata = {}

        o = get_origin(field_type)
        if o is dict:
            dict_args = get_args(field_type)
            if dict_args:
                item_type: type = dict_args[1]

                input_pins, output_pins, _generics = _get_pins(
                    item_type, port_name=port_name, block_type=block_type
                )
                if len(input_pins) or len(output_pins):
                    generics.update(_generics)

                    if dict_args[0] is not str:
                        raise TypeError("Port dictionaries must have str keys")

                    port_interface = PortInterface(
                        metadata=metadata,
                        inputs=input_pins,
                        outputs=output_pins,
                        type=PortType.DICTIONARY,
                        is_function=False,
                    )
                    ports[port_name] = port_interface

                    continue

        elif o is list:
            list_args = get_args(field_type)
            if list_args:
                item_type: type = list_args[0]

                input_pins, output_pins, _generics = _get_pins(
                    item_type, port_name=port_name, block_type=block_type
                )
                if len(input_pins) or len(output_pins):
                    generics.update(_generics)
                    port_interface = PortInterface(
                        metadata=metadata,
                        inputs=input_pins,
                        outputs=output_pins,
                        type=PortType.LIST,
                        is_function=False,
                    )
                    ports[port_name] = port_interface

                    continue

        input_pins, output_pins, _generics = _get_pins(
            port_annotation, port_name=port_name, block_type=block_type
        )
        if len(input_pins) or len(output_pins):
            generics.update(_generics)
            port_interface = PortInterface(
                metadata=metadata,
                inputs=input_pins,
                outputs=output_pins,
                type=PortType.SINGLE,
                is_function=False,
            )
            ports[port_name] = port_interface
        else:
            s = _get_state_from_metadata(
                field_type=port_annotation,
                field_name=port_name,
                block_type=block_type,
            )
            if s:
                state[port_name] = s

    for generic_name, generic_schema in generics.items():
        type_adapter = TypeAdapter(dict[str, Any])
        block_type._set_input_pin_type_adapter(generic_name, "", type_adapter)

        ports[generic_name] = PortInterface(
            metadata={},
            inputs={
                "": InputPinInterface(
                    metadata={"generic": True, "hidden": True},
                    sticky=True,
                    json_schema=type_adapter.json_schema(),
                    generics={},
                    type=PinType.SINGLE,
                    required=False,
                    default=generic_schema.json_schema(),
                    channel=False,
                    virtual=False,
                )
            },
            outputs={},
            type=PortType.SINGLE,
            is_function=False,
        )

    return PortsAndState(ports, state)


class Input(BaseModel):
    sticky: bool = False


class State:
    def __init__(
        self,
        step_id: str | None = None,
        input_ids: list[str] | None = None,
    ):
        self.step_id = step_id
        self.input_ids = input_ids


class BlockControlMessage(enum.Enum):
    DONE = "Done"


block_messages: contextvars.ContextVar[
    asyncio.queues.Queue[BlockRunMessage | BlockControlMessage]
] = contextvars.ContextVar("block_messages")


class OutputChannel(Generic[T]):
    def __init__(self, pin: BlockPinRef):
        self.pin = pin

    def send(self, value: T):
        messages = block_messages.get()
        messages.put_nowait(
            BlockRunMessage(
                outputs=[
                    OutputValue(
                        source=self.pin,
                        value=OutputChannelMessage(
                            data=value,
                            event=ChannelEvent.DATA,
                        ),
                    )
                ],
                inputs=[],
                redirects=[],
                states=[],
            )
        )

    def close(self):
        messages = block_messages.get()
        messages.put_nowait(
            BlockRunMessage(
                outputs=[
                    OutputValue(
                        source=self.pin,
                        value=OutputChannelMessage(
                            data=None,
                            event=ChannelEvent.CLOSE,
                        ),
                    )
                ],
                inputs=[],
                redirects=[],
                states=[],
            )
        )


class Output(Generic[T]):
    def __init__(self, pin: BlockPinRef):
        self.pin = pin

    def send(self, value: T):
        messages = block_messages.get()
        messages.put_nowait(
            BlockRunMessage(
                outputs=[
                    OutputValue(
                        source=self.pin,
                        value=value,
                    )
                ],
                inputs=[],
                redirects=[],
                states=[],
            )
        )


class BlockError(Exception):
    def __init__(self, message: str, data: Any = None):
        self.message = message
        self.data = data

    def __str__(self):
        return f"BlockError: {BlockErrorModel(message=self.message, data=self.data)}"


class ReadOnlyDict(Mapping):
    def __init__(self, data):
        self._data = data

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)


class BlockSet:
    def __init__(self):
        self._blocks: dict[str, dict[str, type[Block]]] = {}

    @property
    def all(self) -> "Mapping[str, dict[str, type[Block]]]":
        return ReadOnlyDict(self._blocks)

    def add(self, block: type["Block"]):
        if block.name not in self._blocks:
            self._blocks[block.name] = {}

        self._blocks[block.name][block.version] = block

    def find(self, name: str, version: str):
        spec = semantic_version.NpmSpec(version)
        if name not in self._blocks:
            return None

        versions = {v.semantic_version: v for v in self._blocks[name].values()}
        best_version = spec.select(versions.keys())

        if best_version is None:
            return None

        return versions[best_version]


class MetaBlock(type):
    def __new__(cls, name, bases, attrs):
        block_type = super().__new__(cls, name, bases, attrs)
        block_type.name = block_type.__name__.split("_")[0]

        return block_type

    def __init__(self, name, bases, attrs):
        super().__init__(name, bases, attrs)

        self.block_class: BlockClass | None = getattr(self, "block_class", None)
        self._scopes: list[BlockScope] | None = getattr(self, "_scopes", None)
        self.metadata: dict[str, Any] = {}
        self.name: str
        self._version: str | None = None
        self._semantic_version: semantic_version.Version | None = None
        self._all_annotations_cache: dict[str, type] | None = None
        self._class_interface: BlockInterface | None = None
        self._input_pin_type_adapters: dict[str, dict[str, TypeAdapter]] = {}
        self._output_pin_type_adapters: dict[str, dict[str, TypeAdapter]] = {}
        self._state_type_adapters: dict[str, TypeAdapter] = {}

    def _set_input_pin_type_adapter(
        self, port: str, pin: str, type_adapter: TypeAdapter
    ):
        if port not in self._input_pin_type_adapters:
            self._input_pin_type_adapters[port] = {}

        self._input_pin_type_adapters[port][pin] = type_adapter

    def _set_output_pin_type_adapter(
        self, port: str, pin: str, type_adapter: TypeAdapter
    ):
        if port not in self._output_pin_type_adapters:
            self._output_pin_type_adapters[port] = {}

        self._output_pin_type_adapters[port][pin] = type_adapter

    def _get_interface(cls):
        if cls._class_interface is None:
            ports, state = _get_ports_and_state(cls)

            for attribute_name in dir(cls):
                attribute = getattr(cls, attribute_name)

                if type(attribute) is Step or type(attribute) is Callback:
                    (inputs, input_adapters, (output, output_adapter), generics) = (
                        _get_function_pins(attribute._fn)
                    )

                    if type(attribute) is Step and output_adapter and output:
                        cls._set_output_pin_type_adapter(
                            attribute_name, attribute._output_name, output_adapter
                        )
                        outputs = {attribute._output_name: output}
                    else:
                        outputs = {}

                    for name, adapter in input_adapters.items():
                        cls._set_input_pin_type_adapter(attribute_name, name, adapter)

                    metadata = attribute.metadata.copy()
                    if type(attribute) is Callback:
                        metadata["callback"] = True
                        metadata["hidden"] = True

                    ports[attribute_name] = PortInterface(
                        metadata=metadata,
                        inputs=inputs,
                        outputs=outputs,
                        type=PortType.SINGLE,
                        is_function=True,
                    )

                    for generic_name, generic_schema in generics.items():
                        type_adapter = TypeAdapter(dict[str, Any])
                        cls._set_input_pin_type_adapter(generic_name, "", type_adapter)

                        ports[generic_name] = PortInterface(
                            metadata={},
                            inputs={
                                "": InputPinInterface(
                                    metadata={"generic": True, "hidden": True},
                                    sticky=True,
                                    json_schema=type_adapter.json_schema(),
                                    generics={},
                                    type=PinType.SINGLE,
                                    required=False,
                                    default=generic_schema.json_schema(),
                                    channel=False,
                                    virtual=False,
                                )
                            },
                            outputs={},
                            type=PortType.SINGLE,
                            is_function=False,
                        )

            annotations = {}
            for base in _get_all_bases(cls):
                base_annotations = getattr(base, "__annotations__", {})
                annotations.update(base_annotations)
            annotations.update(**cls.__annotations__)

            for port_name, port_annotation in annotations.items():
                port_metadata = getattr(port_annotation, "__metadata__", [])
                if len(port_metadata):
                    field_type = get_args(port_annotation)[0]
                    metadata = first(
                        [a.data for a in port_metadata if isinstance(a, Metadata)], {}
                    )
                else:
                    metadata = {}
                    field_type = port_annotation

                origin = get_origin(field_type)

                if origin is GenericSchema:
                    type_var = get_args(field_type)[0]
                    generic_name = type_var.__name__
                    if generic_name in ports:
                        ports[port_name] = ports[generic_name]
                        del ports[generic_name]

                    if port_name in ports:
                        has_default, default = _get_default(cls, port_name)
                        if has_default:
                            ports[port_name].inputs[""].default = default

                        ports[port_name].inputs[""].metadata["hidden"] = False
                        ports[port_name].inputs[""].metadata["config"] = True
                        ports[port_name].inputs[""].metadata.update(metadata)

                        cls._input_pin_type_adapters[port_name] = {
                            "": cls._input_pin_type_adapters[generic_name][""]
                        }
                        del cls._input_pin_type_adapters[generic_name]

                        for port in ports.values():
                            for pin in list(port.inputs.values()) + list(
                                port.outputs.values()
                            ):
                                for g, pin_ref in pin.generics.items():
                                    if (
                                        g == generic_name
                                        and pin_ref.port == generic_name
                                        and pin_ref.pin == ""
                                    ):
                                        pin.generics[g] = BlockPinRef(
                                            port=port_name, pin=""
                                        )

            cls._class_interface = BlockInterface(
                scopes=cls._scopes,
                metadata=cls.metadata,
                ports=ports,
                state=state,
                block_class=cls.block_class,
            )

        return cls._class_interface

    @property
    def semantic_version(cls):
        if not cls._semantic_version:
            version_str = (
                cls._version or ".".join(cls.__name__.split("_")[1:]) or "1.0.0"
            )
            cls._semantic_version = semantic_version.Version.coerce(version_str)

        return cls._semantic_version

    @property
    def version(cls):
        return str(cls.semantic_version)

    @property
    def _all_annotations(cls):
        if not cls._all_annotations_cache:
            cls._all_annotations_cache = {}
            for base in _get_all_bases(cls):
                base_annotations = getattr(base, "__annotations__", {})
                cls._all_annotations_cache.update(base_annotations)
            cls._all_annotations_cache.update(**cls.__annotations__)

        return cls._all_annotations_cache


def _set_input_pin_value_on_port(
    port: Any,
    pin_name: str,
    pin_index: str | None,
    pin_interface: InputPinInterface,
    value: Any,
):
    if pin_interface.type == PinType.SINGLE:
        setattr(port, pin_name, value)

    elif pin_interface.type == PinType.LIST:
        try:
            pin_index_int = int(pin_index or "")
        except ValueError:
            raise ValueError("Indexes on list Pins must be valid integers")
        pin_list = getattr(port, pin_name, None)
        if not pin_list:
            pin_list = []
            setattr(port, pin_name, pin_list)

        if len(pin_list) < pin_index_int:
            pin_list.extend([None] * (pin_index_int - len(pin_list)))
            pin_list.append(value)
        elif len(pin_list) == pin_index_int:
            pin_list.append(value)
        else:
            pin_list[pin_index_int] = value

    elif pin_interface.type == PinType.DICTIONARY:
        pin_dict = getattr(port, pin_name, None)
        if not pin_dict:
            pin_dict = {}
            setattr(port, pin_name, pin_dict)

        pin_dict[pin_index] = value


class Block(metaclass=MetaBlock):
    error: Annotated[Output[BlockErrorModel], Metadata(hidden=True)]

    def __init__(self):
        self._interface = self.interface()

        self._has_run = False
        self._messages: list[BlockRunMessage] = []
        self._dynamic_ports: dict[str, list[str]] = {}
        self._dynamic_inputs: list[tuple[tuple[str, str], tuple[str, str]]] = []
        self._dynamic_outputs: list[tuple[tuple[str, str], tuple[str, str]]] = []
        self._tools: list[Tool] = []

        for attribute_name in dir(self):
            attribute = getattr(self, attribute_name)

            if _issubclass(type(attribute), BlockFunction):
                new_attribute = attribute.create(self)
                setattr(self, attribute_name, new_attribute)

        self._create_all_ports()

    def _run_function(self, name: str):
        function = getattr(self, name, None)
        if function is None:
            raise ValueError(f"Could not find function '{name}'")

        if not isinstance(function, BlockFunction):
            raise ValueError(f"'{name}' is not a BlockFunction")

        return function._run()

    def _load(
        self,
        context: FlowContext | None = None,
        state: list[StateValue] | None = None,
        inputs: list[InputValue] | None = None,
        dynamic_ports: list[str] | None = None,
        dynamic_output_pins: list[BlockPinRef] | None = None,
        dynamic_input_pins: list[BlockPinRef] | None = None,
    ):
        if (
            (dynamic_input_pins and len(dynamic_input_pins))
            or (dynamic_output_pins and len(dynamic_output_pins))
            or (dynamic_ports and len(dynamic_ports))
        ):
            self._create_all_ports(
                dynamic_ports, dynamic_input_pins, dynamic_output_pins
            )

        if context:
            self._set_context(context)

        if state:
            self._set_state(state)

        if inputs:
            self._set_inputs(inputs)

    def get_messages(self):
        return copy.copy(self._messages)

    @classmethod
    def interface(cls):
        return cls._get_interface().model_copy(deep=True)

    def _create_all_ports(
        self,
        dynamic_ports: list[str] | None = None,
        dynamic_input_pins: list[BlockPinRef] | None = None,
        dynamic_output_pins: list[BlockPinRef] | None = None,
    ):
        for port_name, port_interface in self.interface().ports.items():
            if (
                port_interface.type == PortType.LIST
                or port_interface.type == PortType.DICTIONARY
            ):
                self._dynamic_ports[port_name] = []

        if dynamic_ports:
            for i in dynamic_ports:
                port_path = i.split(".")
                port_name = port_path[0]
                port_index = port_path[1] if len(port_path) == 2 else ""

                if port_index:
                    self._dynamic_ports[port_name].append(port_index)

        if dynamic_input_pins:
            for i in dynamic_input_pins:
                port_path = i.port.split(".")
                pin_path = i.port.split(".")

                port_name = port_path[0]
                port_index = port_path[1] if len(port_path) == 2 else ""

                self._dynamic_inputs.append(
                    (
                        (port_name, port_index),
                        (pin_path[0], pin_path[1] if len(pin_path) == 2 else ""),
                    )
                )

        if dynamic_output_pins:
            for i in dynamic_output_pins:
                port_path = i.port.split(".")
                pin_path = i.port.split(".")

                port_name = port_path[0]
                port_index = port_path[1] if len(port_path) == 2 else ""

                self._dynamic_outputs.append(
                    (
                        (port_name, port_index),
                        (pin_path[0], pin_path[1] if len(pin_path) == 2 else ""),
                    )
                )

        for port_name, port_interface in self.interface().ports.items():
            if port_interface.type == PortType.SINGLE:
                setattr(
                    self,
                    port_name,
                    self._create_port(port_name, ""),
                )

            elif port_interface.type == PortType.LIST:
                port_indexes = [
                    int(port_index) for port_index in self._dynamic_ports[port_name]
                ]

                port_list: list[Any] = [None] * (max(port_indexes, default=-1) + 1)
                for port_index in port_indexes:
                    port_list[port_index] = self._create_port(
                        port_name, str(port_index)
                    )
                setattr(self, port_name, port_list)

            elif port_interface.type == PortType.DICTIONARY:
                port_dict = {
                    port_index: self._create_port(port_name, port_index)
                    for port_index in self._dynamic_ports[port_name]
                }
                setattr(self, port_name, port_dict)

    def _set_context(self, context: FlowContext): ...

    def _set_state(self, state: list[StateValue]):
        for s in state:
            adapter = self.__class__._state_type_adapters[s.state]
            try:
                value = adapter.validate_python(s.value)
            except ValidationError:
                value = s.value

            setattr(self, s.state, value)

    def _set_inputs(self, inputs: list[InputValue]):
        for input_value in inputs:
            port_path = input_value.target.port.split(".")
            port_name = port_path[0]
            if len(port_path) > 1:
                port_index = port_path[1]
            else:
                port_index = ""

            pin_path = input_value.target.pin.split(".")
            pin_name = pin_path[0]
            if len(pin_path) > 1:
                pin_index = pin_path[1]
            else:
                pin_index = ""

            adapter = self.__class__._input_pin_type_adapters[port_name][pin_name]

            try:
                value = adapter.validate_python(input_value.value)
            except ValidationError:
                value = input_value.value

            if (
                port_name in self.interface().ports
                and pin_name in self.interface().ports[port_name].inputs
            ):
                port_interface = self.interface().ports[port_name]
                pin_interface = port_interface.inputs[pin_name]

                if port_interface.is_function:
                    port: BlockFunction = getattr(self, port_name)
                    if pin_name not in port._pending_inputs:
                        port._pending_inputs[pin_name] = {}
                    port._pending_inputs[pin_name][pin_index] = value

                else:
                    if pin_name == "":
                        assert pin_interface.type == PinType.SINGLE

                        if port_interface.type == PortType.SINGLE:
                            setattr(self, port_name, value)

                        elif port_interface.type == PortType.LIST:
                            try:
                                port_index_int = int(port_index)
                            except ValueError:
                                raise ValueError(
                                    "Indexes on list Ports must be valid integers"
                                )

                            port_list: list[Any] = getattr(self, port_name)

                            if len(port_list) < port_index_int:
                                port_list.extend(
                                    [None] * (port_index_int - len(port_list))
                                )
                                port_list.append(value)
                            elif len(port_list) == port_index_int:
                                port_list.append(value)
                            else:
                                port_list[port_index_int] = value

                        elif port_interface.type == PortType.DICTIONARY:
                            port_dict: dict[str, Any] = getattr(self, port_name)
                            port_dict[port_index] = value
                    else:
                        if port_interface.type == PortType.SINGLE:
                            port = getattr(self, port_name)

                        elif port_interface.type == PortType.LIST:
                            try:
                                port_index_int = int(port_index)
                            except ValueError:
                                raise ValueError(
                                    "Indexes on list Ports must be valid integers"
                                )

                            port_list = getattr(self, port_name)
                            port = port_list[port_index_int]

                        elif port_interface.type == PortType.DICTIONARY:
                            port_dict = getattr(self, port_name)
                            port = port_dict[port_index]

                        _set_input_pin_value_on_port(
                            port,
                            pin_name=pin_name,
                            pin_index=pin_index,
                            pin_interface=pin_interface,
                            value=value,
                        )

    def _create_port(
        self,
        port_name: str,
        port_index: str,
    ) -> Any:
        port_id = port_name if not port_index else f"{port_name}.{port_index}"

        port_interface = self.interface().ports[port_name]
        dynamic_inputs: list[tuple[str, str]] = []
        for (_port_name, _port_index), (
            _input_name,
            _input_index,
        ) in self._dynamic_inputs:
            if _port_name == port_name and _port_index == port_index:
                dynamic_inputs.append((_input_name, _input_index))

        dynamic_outputs: list[tuple[str, str]] = []
        for (_port_name, _port_index), (
            _output_name,
            _output_index,
        ) in self._dynamic_outputs:
            if _port_name == port_name and _port_index == port_index:
                dynamic_outputs.append((_output_name, _output_index))

        if len(port_interface.inputs) + len(port_interface.outputs) == 1 and (
            "" in port_interface.inputs or "" in port_interface.outputs
        ):
            if "" in port_interface.inputs:
                input_interface = port_interface.inputs[""]
                type_adapter = self._input_pin_type_adapters[port_name][""]
                return (
                    None
                    if input_interface.default is None
                    else type_adapter.validate_python(input_interface.default)
                )
            elif "" in port_interface.outputs:
                if port_interface.outputs[""].channel:
                    return OutputChannel(BlockPinRef(port=port_id, pin=""))
                else:
                    return Output(BlockPinRef(port=port_id, pin=""))

        tool_port = None
        if port_interface.is_function:
            port = getattr(self, port_name)
        else:
            annotation = self.__class__._all_annotations[port_name]
            if port_interface.type == PortType.SINGLE:
                port_type = annotation
            else:
                if get_origin(annotation) == Annotated:
                    annotation = get_args(annotation)[0]

                if port_interface.type == PortType.LIST:
                    port_type = get_args(annotation)[0]
                elif port_interface.type == PortType.DICTIONARY:
                    port_type = get_args(annotation)[1]

            if _issubclass(port_type, Tool):
                port = port_type(port_name=port_id, input_names=[])
                self._tools.append(port)
                tool_port = cast(Tool, port)
            else:
                port = port_type()

        for input_name, input_interface in port_interface.inputs.items():
            type_adapter = self._input_pin_type_adapters[port_name][input_name]

            if input_interface.type == PinType.SINGLE:
                setattr(
                    port,
                    input_name,
                    None
                    if input_interface.default is None
                    else type_adapter.validate_python(input_interface.default),
                )

            elif input_interface.type == PinType.LIST:
                _dynamic_inputs = [
                    int(index)
                    for _input_name, index in dynamic_inputs
                    if _input_name == input_name
                ]
                inputs = [None] * (max(_dynamic_inputs, default=-1) + 1)

                for index in _dynamic_inputs:
                    inputs[index] = (
                        None
                        if input_interface.default is None
                        else type_adapter.validate_python(input_interface.default)
                    )

                setattr(port, input_name, inputs)

            elif input_interface.type == PinType.DICTIONARY:
                input_dict = {
                    index: None
                    if input_interface.default is None
                    else type_adapter.validate_python(input_interface.default)
                    for _input_name, index in dynamic_inputs
                    if _input_name == input_name
                }

                setattr(port, input_name, input_dict)

        for output_name, output_interface in port_interface.outputs.items():
            if output_interface.type == PinType.SINGLE:
                if output_interface.channel:
                    output = OutputChannel(BlockPinRef(port=port_id, pin=output_name))
                else:
                    output = Output(BlockPinRef(port=port_id, pin=output_name))

                setattr(port, output_name, output)

                if tool_port:
                    tool_port.output_names.append(output_name)

            elif output_interface.type == PinType.LIST:
                _dynamic_outputs = [
                    int(index)
                    for _output_name, index in dynamic_outputs
                    if _output_name == output_name
                ]
                outputs: list[None | Output | OutputChannel] = [None] * (
                    max(_dynamic_outputs, default=-1) + 1
                )

                for index in _dynamic_outputs:
                    if output_interface.channel:
                        outputs[index] = OutputChannel(
                            BlockPinRef(port=port_id, pin=output_name)
                        )
                    else:
                        outputs[index] = Output(
                            BlockPinRef(port=port_id, pin=output_name)
                        )

                setattr(port, output_name, outputs)

                if tool_port:
                    tool_port.output_names.extend(
                        [f"{output_name}.{i}" for i, _ in enumerate(outputs)]
                    )

            elif output_interface.type == PinType.DICTIONARY:
                output_dict: dict[str, Output | OutputChannel] = {
                    index: OutputChannel(BlockPinRef(port=port_id, pin=output_name))
                    if output_interface.channel
                    else Output(BlockPinRef(port=port_id, pin=output_name))
                    for _output_name, index in dynamic_outputs
                    if _output_name == output_name
                }
                setattr(port, output_name, output_dict)

                if tool_port:
                    tool_port.output_names.extend(
                        [f"{output_name}.{i}" for i in output_dict.keys()]
                    )

        return port


class WorkSpaceBlock(Block):
    workspace: SmartSpaceWorkspace
    message_history: list[ThreadMessage]
    _scopes = [BlockScope.WORKSPACE]

    def _set_context(self, context: FlowContext):
        assert context.workspace is not None, "Workspace is None in a WorkSpaceBlock"
        assert context.message_history is not None, (
            "Workspace is None in a WorkSpaceBlock"
        )

        self.workspace = context.workspace
        self.message_history = context.message_history


class OperatorBlock(Block):
    block_class = BlockClass.OPERATOR


class DummyToolValue: ...


class CallbackCall(NamedTuple):
    name: str
    other_params: dict[str, Any]
    dummy_value_param: str


class ToolCall(Generic[R]):
    def __init__(self, port_name: str, outputs: list[OutputValue]):
        self.port_name = port_name
        self.outputs = outputs
        self.inputs: list[InputValue] = []
        self.redirects: list[PinRedirect] = []

    def then(
        self,
        callback: Callable[[R], CallbackCall],
    ) -> "ToolCall[R]":
        callback_name, other_params, dummy_value_param = callback(
            cast(R, DummyToolValue())
        )

        for name, value in other_params.items():
            self.inputs.append(
                InputValue(
                    target=BlockPinRef(
                        port=callback_name,
                        pin=name,
                    ),
                    value=value,
                )
            )

        self.redirects.append(
            PinRedirect(
                source=BlockPinRef(
                    port=self.port_name,
                    pin="return",
                ),
                target=BlockPinRef(
                    port=callback_name,
                    pin=dummy_value_param,
                ),
            )
        )

        return self

    def __await__(self):
        messages = block_messages.get()

        messages.put_nowait(
            BlockRunMessage(
                outputs=self.outputs,
                inputs=self.inputs,
                redirects=self.redirects,
                states=[],
            )
        )

        yield


class Tool(Generic[P, T], abc.ABC):
    metadata: ClassVar[dict] = {}

    def __init__(self, port_name: str, input_names: list[str]):
        self.port_name = port_name
        self.output_names = input_names

    @abc.abstractmethod
    def run(self, *args: P.args, **kwargs: P.kwargs) -> T: ...

    def call(self, *args: P.args, **kwargs: P.kwargs) -> ToolCall[T]:
        s = inspect.signature(self.__class__.run)
        binding = s.bind(self, *args, **kwargs)
        binding.apply_defaults()

        single_outputs: list[OutputValue] = []
        list_outputs: list[OutputValue] = []
        dictionary_outputs: list[OutputValue] = []

        for name, p in s.parameters.items():
            if name == "self":
                continue

            value = binding.arguments[name]

            if p.kind == p.POSITIONAL_OR_KEYWORD or p.kind == p.KEYWORD_ONLY:
                single_outputs.append(
                    OutputValue(
                        source=BlockPinRef(
                            port=self.port_name,
                            pin=name,
                        ),
                        value=OutputChannelMessage(
                            data=value,
                            event=ChannelEvent.DATA,
                        ),
                    )
                )
            elif p.kind == p.VAR_POSITIONAL:
                for i, v in enumerate(value):
                    list_outputs.append(
                        OutputValue(
                            source=BlockPinRef(
                                port=self.port_name,
                                pin=f"{name}.{i}",
                            ),
                            value=OutputChannelMessage(
                                data=v,
                                event=ChannelEvent.DATA,
                            ),
                        )
                    )
            elif p.kind == p.VAR_KEYWORD:
                value = cast(dict[str, Any], value)
                for i, v in value.items():
                    dictionary_outputs.append(
                        OutputValue(
                            source=BlockPinRef(
                                port=self.port_name,
                                pin=f"{name}.{i}",
                            ),
                            value=OutputChannelMessage(
                                data=v,
                                event=ChannelEvent.DATA,
                            ),
                        )
                    )

        all_outputs = single_outputs + list_outputs + dictionary_outputs

        return ToolCall(port_name=self.port_name, outputs=all_outputs)


class BlockFunctionCall:
    def __init__(
        self,
        values: asyncio.queues.Queue[BlockRunMessage | BlockControlMessage],
        step: Awaitable,
    ):
        self.values = values
        self.step = step
        self.result: Any = None

    def _on_done(self, task: asyncio.Task):
        self.values.put_nowait(BlockControlMessage.DONE)
        if not task.cancelled() and not task.exception():
            self.result = task.result()

    def __aiter__(self):
        self.step_future = asyncio.tasks.ensure_future(self.step)
        self.step_future.add_done_callback(self._on_done)

        return self

    async def __anext__(self):
        value = await self.values.get()

        if isinstance(value, BlockControlMessage):
            if value == BlockControlMessage.DONE:
                exc = self.step_future.exception()
                if exc:
                    raise exc

                raise StopAsyncIteration
            else:
                raise ValueError(f"Unexpected BlockControlMessage {value}")
        elif isinstance(value, BlockRunMessage):
            return value
        else:
            raise ValueError(f"Unexpected BlockMessage {value}")


class BlockFunction(Generic[B, P, T]):
    def __init__(
        self,
        fn: Callable[Concatenate[B, P], Awaitable[T]],
        output_name: str | None = None,
    ):
        self.name = fn.__name__
        self._fn = fn
        self._output_name = output_name or ""
        self.metadata: dict = {}
        self._block: B
        self._pending_inputs: dict[str, dict[str, Any]] = {}

    def create(self, block: Block) -> "BlockFunction":
        if isinstance(self, Callback):
            instance = Callback(self._fn)
        else:
            instance = Step(self._fn, self._output_name)

        instance._block = block
        return instance

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        call = await self._call_inner(*args, **kwargs)

        async for m in call:
            self._block._messages.append(m)

        return call.result

    async def _run(self):
        s = inspect.signature(self._fn)

        positional_inputs: list[Any] = []
        var_positional_inputs: list[Any] = []
        keyword_inputs: dict[str, Any] = {}

        for name, p in s.parameters.items():
            if name == "self" or name not in self._pending_inputs:
                continue

            values = self._pending_inputs[name]

            if p.kind == p.POSITIONAL_OR_KEYWORD or p.kind == p.KEYWORD_ONLY:
                if "" in values:
                    positional_inputs.append(
                        values[""]
                    )  # s.parameters is an ordered dict
            elif p.kind == p.VAR_POSITIONAL:
                indexed_values = {int(index): value for index, value in values.items()}
                var_positional_inputs = [None] * (
                    max(indexed_values.keys(), default=-1) + 1
                )
                for i, v in indexed_values.items():
                    var_positional_inputs[i] = v

            elif p.kind == p.VAR_KEYWORD:
                keyword_inputs.update(values)

        return await self._call_inner(
            *tuple(positional_inputs + var_positional_inputs),
            **keyword_inputs,
        )

    async def _call_inner(self, *args: P.args, **kwargs: P.kwargs) -> BlockFunctionCall:
        if self._block._has_run:
            raise BlockError(
                message="Block has already run a function. Each instance of a block can only run once",
                data={"function_name": self.name},
            )

        self._block._has_run = True

        messages: asyncio.queues.Queue[BlockRunMessage | BlockControlMessage] = (
            asyncio.queues.Queue()
        )
        block_messages.set(messages)

        async def _inner() -> T:
            result = await self._fn(
                self._block,
                *args,
                **kwargs,
            )

            outputs: list[OutputValue] = []
            states: list[StateValue] = []

            s = inspect.signature(self._fn)
            if s.return_annotation is not inspect._empty:
                outputs = [
                    OutputValue(
                        source=BlockPinRef(port=self.name, pin=self._output_name),
                        value=result,
                    )
                ]

            for state_name in self._block._interface.state.keys():
                state_value = getattr(self._block, state_name, None)
                states.append(
                    StateValue(
                        state=state_name,
                        value=state_value,
                    )
                )

            messages.put_nowait(
                BlockRunMessage(
                    outputs=outputs,
                    inputs=[],
                    redirects=[],
                    states=states,
                )
            )

            tool_close_outputs = [
                OutputValue(
                    source=BlockPinRef(
                        port=tool.port_name,
                        pin=tool.output_names[0],
                    ),
                    value=OutputChannelMessage(
                        data=None,
                        event=ChannelEvent.CLOSE,
                    ),
                )
                for tool in self._block._tools
            ]

            messages.put_nowait(
                BlockRunMessage(
                    outputs=tool_close_outputs,
                    inputs=[],
                    redirects=[],
                    states=[],
                )
            )

            return result

        return BlockFunctionCall(
            messages,
            _inner(),
        )


class Step(BlockFunction[B, P, T]):
    def __init__(
        self,
        fn: Callable[Concatenate[B, P], Awaitable[T]],
        output_name: str | None = None,
    ):
        super().__init__(fn)
        self._output_name = output_name or ""


class Callback(BlockFunction[B, P, None]):
    def __init__(
        self,
        fn: Callable[Concatenate[B, P], Awaitable[None]],
    ):
        super().__init__(fn, None)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> CallbackCall:
        values = inspect.getcallargs(self._fn, cast(Block, None), *args, **kwargs)

        tool_result_param = ""
        direct_params: dict[str, Any] = {}

        for arg_name, value in values.items():
            if isinstance(value, DummyToolValue):
                tool_result_param = arg_name
            elif arg_name != "self":
                direct_params[arg_name] = value

        return CallbackCall(
            name=self.name,
            other_params=direct_params,
            dummy_value_param=tool_result_param,
        )


def metadata(
    description: str | None = None,  # short description, for tooltips and things
    display_type: InputDisplayType | None = None,  # type of display
    documentation: str | None = None,  # long description
    category: BlockCategory | dict[str, Any] | None = None,
    icon: str | None = None,  # fontawesome 5 icon name
    obsolete: bool | None = None,
    label: str | None = None,  # searchable label for the block
    deprecated_reason: str | None = None,
    use_instead: str | None = None,  # use this block instead of the current one
    **kwargs,
):
    if description is not None:
        kwargs["description"] = description
    if documentation is not None:
        kwargs["documentation"] = documentation
    if category is not None:
        kwargs["category"] = category
    if icon is not None:
        kwargs["icon"] = icon
    if obsolete is not None:
        kwargs["obsolete"] = obsolete
    if label is not None:
        kwargs["label"] = label
    if use_instead is not None:
        kwargs["use_instead"] = use_instead
    if deprecated_reason is not None:
        kwargs["deprecated_reason"] = deprecated_reason

    def _inner(cls):
        setattr(cls, "metadata", kwargs)
        return cls

    return _inner


def version(version: str):
    def inner(cls: type[B]) -> type[B]:
        cls._version = version
        return cls

    return inner


def step(
    output_name: str | None = None,
) -> Callable[[Callable[Concatenate[B, P], Awaitable[T]]], Step[B, P, T]]:
    def step_decorator(fn: Callable[Concatenate[B, P], Awaitable[T]]) -> Step[B, P, T]:
        if not inspect.iscoroutinefunction(fn):
            raise TypeError(f"Steps must be async and step {fn.__name__} is not")

        return Step[B, P, T](fn, output_name=output_name)

    return step_decorator


def callback() -> Callable[[Callable[Concatenate[B, P], Awaitable]], Callback[B, P]]:
    def callback_decorator(
        fn: Callable[Concatenate[B, P], Awaitable[None]],
    ) -> Callback[B, P]:
        if not inspect.iscoroutinefunction(fn):
            raise TypeError(f"Callbacks must be async and step {fn.__name__} is not")

        return Callback[B, P](fn)

    return callback_decorator


UserMessageT = TypeVar("UserMessageT")


class User(WorkSpaceBlock, Generic[UserMessageT]):
    schema: GenericSchema[UserMessageT] = GenericSchema({"type": "string"})
    response: Output[UserMessageT]

    @step()
    async def ask(self, message: str): ...

    @callback()
    async def handle_response(self, response: UserMessageT):
        self.response.send(response)



import asyncio
from collections import deque
from typing import Annotated, List, Set
from urllib.parse import urljoin, urlparse

import httpx
from pydantic import BaseModel

from smartspace.core import Block, Config, Output, metadata, step
from smartspace.enums import BlockCategory


class WebsiteDetails(BaseModel):
    title: str
    url: str
    content: str


@metadata(
    description="Scrapes the content of a website. Returns both the raw content and the content with metadata.",
    category=BlockCategory.MISC,
    icon="fa-globe",
    label="website scraper, web crawler, content extractor, web harvester, site parser",
)
class WebsiteScraper(Block):
    website_content: Output[list[str]]
    website_details: Output[list[WebsiteDetails]]
    page_limit: Annotated[int, Config()] = 3

    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3"
    }

    @step()
    async def scrape_website(self, base_url: str):
        from bs4 import BeautifulSoup

        if not base_url:
            self.website_content.send([""])
            return

        if not urlparse(base_url).scheme:
            base_url = f"https://{base_url}"

        visited: Set[str] = set()
        pages_to_visit = deque([base_url])
        scraped_content: List[WebsiteDetails] = []

        async def fetch_page(url: str):
            try:
                async with httpx.AsyncClient(timeout=10) as client:
                    response = await client.get(url, headers=self.headers)
                    if response.status_code == 301:
                        url = response.headers["Location"]
                        response = await client.get(url, headers=self.headers)
                    response.raise_for_status()
                    return response
            except httpx.HTTPError as e:
                print(f"Failed to fetch {url}: {e}")
                return None

        async def process_page(url: str):
            if url in visited:
                return

            response = await fetch_page(url)
            if not response:
                return

            visited.add(url)
            soup = BeautifulSoup(response.text, "html.parser")

            # Extract the main content
            for script in soup(["script", "style"]):
                script.decompose()  # Remove unnecessary tags

            text = soup.get_text(separator="\n")
            lines = [line.strip() for line in text.splitlines()]
            text = "\n".join(line for line in lines if line)
            title = soup.title.string if soup.title and soup.title.string else ""
            scraped_content.append(WebsiteDetails(title=title, url=url, content=text))

            # Find and queue additional pages to visit
            for link_tag in soup.find_all("a", href=True):
                link = link_tag["href"]
                full_url = urljoin(base_url, link)
                # Stay within the same domain
                if urlparse(full_url).netloc == urlparse(base_url).netloc:
                    if full_url not in visited and full_url not in pages_to_visit:
                        pages_to_visit.append(full_url)

        # process first page
        await process_page(base_url)
        tasks = []
        while (
            pages_to_visit
            # limit the number of pages to visit
            and len(visited) < self.page_limit
        ):
            url = pages_to_visit.popleft()
            if url not in visited:
                tasks.append(asyncio.create_task(process_page(url)))
            # run when we have 5 tasks or we have reached the page limit
            if (len(tasks) + len(visited) == self.page_limit) or len(tasks) >= 5:
                await asyncio.gather(*tasks)
                tasks = []
                await asyncio.sleep(1)  # Rate limiting

        if not len(scraped_content):
            self.website_content.send([""])
            self.website_details.send([])
            return

        self.website_content.send([webpage.content for webpage in scraped_content])
        self.website_details.send(scraped_content)



import json
from typing import Annotated, Optional

import typer

from smartspace.cli import blocks
from smartspace.cli.config import SmartSpaceConfig

app = typer.Typer()
app.add_typer(blocks.app, name="blocks")


@app.command()
def config(
    apiUrl: Annotated[Optional[str], typer.Option("--api-url")] = None,
    tenantId: Annotated[Optional[str], typer.Option("--tenant-id")] = None,
    clientId: Annotated[Optional[str], typer.Option("--client-id")] = None,
):
    from smartspace.cli.config import load_config, save_config

    current_config = load_config()
    new_config = SmartSpaceConfig(
        config_api_url=current_config["config_api_url"] if apiUrl is None else apiUrl,
        tenant_id=current_config["tenant_id"] if tenantId is None else tenantId,
        client_id=current_config["client_id"] if clientId is None else clientId,
    )

    if apiUrl or tenantId or clientId:
        save_config(new_config)
        print("Config updated")

    print(json.dumps(new_config, indent=2))


@app.command()
def login(deviceCode: Annotated[bool, typer.Option("--device-code")] = False):
    from smartspace.cli import auth

    auth.login(deviceCode=deviceCode)


def cli():
    app()


if __name__ == "__main__":
    cli()



import json
from typing import Annotated, Any

from smartspace.core import Block, Config, Metadata, metadata, step
from smartspace.enums import BlockCategory, InputDisplayType


@metadata(
    category=BlockCategory.FUNCTION,
    description="""
    A block that takes a Jinja2 template string, fills the template,
    and then parses the result into a JSON object.
    """,
    icon="fa-code",
    label="template object, JSON templating, dynamic JSON, structured template, object generation",
)
class TemplatedObject(Block):
    templated_json: Annotated[
        str,
        Config(),
        Metadata(
            display_type=InputDisplayType.TEMPLATEOBJECT,
            description="The Jinja2 template string that is formatted and parsed to JSON",
        ),
    ]

    @step(output_name="json")
    async def add_files(
        self,
        **inputs: Annotated[
            Any, Metadata(description="Objects passed to the Jinja2 template")
        ],
    ) -> dict[str, Any]:
        """
        Render the templated JSON from the Jinja2 template (self.templated_json),
        using the provided inputs, and then parse the result as JSON.
        """

        from jinja2 import BaseLoader, Environment, TemplateError

        try:
            env = Environment(loader=BaseLoader())
            template = env.from_string(self.templated_json)

            # Convert each input value into an AutoJson wrapper.
            # This allows dot-notation (e.g. person.sports) to become JSON automatically.
            wrapped_inputs = {k: wrap_auto_json(v) for k, v in inputs.items()}

            rendered_json = template.render(**wrapped_inputs)

            parsed_json = json.loads(rendered_json)
            return parsed_json

        except TemplateError as e:
            raise ValueError(f"Error in rendering Jinja2 template: {e}")
        except json.JSONDecodeError as e:
            raise ValueError(
                f"Error in parsing rendered template to JSON: {e}\n"
                f"Rendered output was:\n{rendered_json}"
            )


class AutoJsonWrapper:
    """
    Base class that ensures that when converted to a string,
    we produce valid JSON of the wrapped data.
    """

    def __init__(self, data):
        self._data = data

    def __str__(self):
        from markupsafe import Markup

        # When Jinja calls str(...) on this object, we dump it as JSON
        # and mark it safe so it doesn't get escaped again.
        return Markup(json.dumps(self._unwrap()))

    def _unwrap(self):
        # By default, just return the underlying data.
        # Subclasses can override if needed.
        return self._data


class AutoJsonDict(AutoJsonWrapper):
    """
    Wraps a dict so that attribute or item lookups return more wrappers.
    """

    def __getitem__(self, key):
        return wrap_auto_json(self._data[key])

    def __getattr__(self, key):
        # If we do person.sports, Jinja calls __getattr__('sports')
        return self.__getitem__(key)

    def _unwrap(self):
        # Recursively produce a normal dict for final json.dumps
        return {k: unwrap_for_json(v) for k, v in self._data.items()}


class AutoJsonList(AutoJsonWrapper):
    """
    Wraps a list so that item lookups return more wrappers.
    """

    def __getitem__(self, idx):
        return wrap_auto_json(self._data[idx])

    def __len__(self):
        return len(self._data)

    def _unwrap(self):
        return [unwrap_for_json(item) for item in self._data]


class AutoJsonScalar(AutoJsonWrapper):
    """
    Wraps a scalar (string, int, float, bool, None)
    """

    # In most cases, the base class behavior is enough.
    # We just define it for clarity.


def wrap_auto_json(value):
    """
    Return an AutoJson wrapper appropriate for the given value.
    """
    if isinstance(value, dict):
        return AutoJsonDict(value)
    elif isinstance(value, list):
        return AutoJsonList(value)
    else:
        # String, int, float, bool, or any other scalar
        return AutoJsonScalar(value)


def unwrap_for_json(wrapper):
    """
    If 'wrapper' is an AutoJsonWrapper, recursively convert it to
    normal Python objects for final JSON serialization. Otherwise
    just return it.
    """
    if isinstance(wrapper, AutoJsonDict):
        return {k: unwrap_for_json(v) for k, v in wrapper._data.items()}
    elif isinstance(wrapper, AutoJsonList):
        return [unwrap_for_json(item) for item in wrapper._data]
    elif isinstance(wrapper, AutoJsonScalar):
        return wrapper._data
    else:
        return wrapper



from unittest.mock import patch

import pytest

from smartspace.blocks.truncate_string import StringTruncator


@pytest.mark.asyncio
async def test_truncate_string_no_truncation_needed():
    truncator = StringTruncator()
    input_string = "This is a short string that doesn't need truncation."

    with patch("smartspace.blocks.truncate_string.encode") as mock_encode:
        mock_encode.return_value = [1] * 50  # Simulate 50 tokens
        result = await truncator.truncate_string(input_string)

    assert result == input_string


@pytest.mark.asyncio
async def test_truncate_string_truncation_needed():
    truncator = StringTruncator()
    input_string = "This is a long string that needs truncation." * 10

    with (
        patch("smartspace.blocks.truncate_string.encode") as mock_encode,
        patch("smartspace.blocks.truncate_string.decode") as mock_decode,
    ):
        mock_encode.return_value = [1] * 150  # Simulate 150 tokens
        mock_decode.return_value = "This is a truncated string."
        result = await truncator.truncate_string(input_string)

    assert result == "This is a truncated string."
    mock_encode.assert_called_once_with(model=truncator.model_name, text=input_string)
    mock_decode.assert_called_once_with(model=truncator.model_name, tokens=[1] * 100)


@pytest.mark.asyncio
async def test_truncate_string_custom_token_limit():
    truncator = StringTruncator()
    truncator.max_token = 50
    input_string = "This string should be truncated to a custom token limit."

    with (
        patch("smartspace.blocks.truncate_string.encode") as mock_encode,
        patch("smartspace.blocks.truncate_string.decode") as mock_decode,
    ):
        mock_encode.return_value = [1] * 75  # Simulate 75 tokens
        mock_decode.return_value = "This string should be truncated to a custom"
        result = await truncator.truncate_string(input_string)

    assert result == "This string should be truncated to a custom"
    mock_decode.assert_called_once_with(model=truncator.model_name, tokens=[1] * 50)


@pytest.mark.asyncio
async def test_truncate_string_custom_model():
    truncator = StringTruncator()
    truncator.model_name = "gpt-4"
    input_string = "This string uses a different model for encoding and decoding."

    with (
        patch("smartspace.blocks.truncate_string.encode") as mock_encode,
        patch("smartspace.blocks.truncate_string.decode") as mock_decode,
    ):
        mock_encode.return_value = [1] * 180  # Simulate 80 tokens
        mock_decode.return_value = "This string uses a different model for encoding"
        result = await truncator.truncate_string(input_string)

    assert result == "This string uses a different model for encoding"
    mock_encode.assert_called_once_with(model="gpt-4", text=input_string)
    mock_decode.assert_called_once_with(model="gpt-4", tokens=[1] * truncator.max_token)


@pytest.mark.asyncio
async def test_truncate_string_empty_input():
    truncator = StringTruncator()
    input_string = ""

    with patch("smartspace.blocks.truncate_string.encode") as mock_encode:
        mock_encode.return_value = []  # Simulate 0 tokens
        result = await truncator.truncate_string(input_string)

    assert result == ""


@pytest.mark.asyncio
async def test_truncate_string_exact_token_limit():
    truncator = StringTruncator()
    input_string = "This string has exactly the token limit."

    with patch("smartspace.blocks.truncate_string.encode") as mock_encode:
        mock_encode.return_value = [1] * 100  # Simulate exactly 100 tokens
        result = await truncator.truncate_string(input_string)

    assert result == input_string



from typing import Annotated, Any, Generic, TypeVar

from smartspace.core import (
    Block,
    Config,
    Output,
    metadata,
    step,
)
from smartspace.enums import BlockCategory
from smartspace.utils.expressions import evaluate_expression, expression_tooltip

ValueT = TypeVar("ValueT")


@metadata(
    category=BlockCategory.FUNCTION,
    description=f"""Evaluates the input value using the configured condition.
{expression_tooltip}""",
    label="conditional logic, boolean check, if-then-else, branching, condition evaluation",
)
class If(Block, Generic[ValueT]):
    condition: Annotated[str, Config()] = "value"
    false: Output[ValueT]
    true: Output[ValueT]

    @step()
    async def create_response(self, value: ValueT):
        if evaluate_expression(self.condition, value):
            self.true.send(value)
        else:
            self.false.send(value)


class SwitchOption:
    condition: Annotated[str, Config()] = "value == "
    output: Output[Any]


@metadata(
    category=BlockCategory.FUNCTION,
    description=f"""Evaluates the input value using the configured condition.
    {expression_tooltip}""",
    label="switch statement, case selection, multi-way branching, conditional routing, expression switch",
)
class Switch(Block, Generic[ValueT]):
    options: dict[str, SwitchOption]

    @step()
    async def create_response(self, value: ValueT):
        try:
            # Collect all matching options based on the condition.
            matching_options = [
                option
                for option in self.options.values()
                if evaluate_expression(option.condition, value)
            ]

            # Check if exactly one option matched.
            if len(matching_options) != 1:
                raise ValueError(
                    f"Switch error: expected exactly one matching option, but found {len(matching_options)}."
                )

            # Send the output from the single matching option.
            matching_options[0].output.send(value)

        except Exception as error:
            # Handle or re-raise the error as needed.
            raise error


@metadata(
    category=BlockCategory.FUNCTION,
    description=f"""Filters a list of items using the configured condition.
{expression_tooltip}""",
    label="list filtering, item selection, data subsetting, conditional filtering, collection filtering",
)
class Filter(Block, Generic[ValueT]):
    condition: Annotated[str, Config()] = "value"

    @step(output_name="items")
    async def create_response(self, items: list[ValueT]) -> list[ValueT]:
        return [i for i in items if evaluate_expression(self.condition, i)]



from enum import Enum
from typing import Annotated, Any, Iterable

import httpx
from pydantic import BaseModel

from smartspace.core import Block, Config, Metadata, metadata, step
from smartspace.enums import BlockCategory


class HTTPMethod(str, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"


class RequestObject(BaseModel):
    method: HTTPMethod | None = None
    url: str = ""
    headers: dict[str, Any] = {}
    query_params: dict[str, Any] = {}
    body: dict[str, Any] | None = None


class ResponseObject(BaseModel):
    content: str | bytes | Iterable[bytes]
    headers: dict[Any, Any]
    body: Any
    status_code: int
    text: str


class HTTPError(Exception):
    def __init__(
        self,
        message: str,
        status_code: int | None = None,
        response: ResponseObject | None = None,
    ):
        self.message = message
        self.status_code = status_code
        self.response = response
        super().__init__(self.message)


@metadata(
    description="Performs HTTP requests such as GET, POST, PUT, DELETE, and more.",
    category=BlockCategory.FUNCTION,
    icon="fa-cloud-download-alt",
    label="HTTP request, web API call, REST client, API request, web service call",
)
class HTTPRequest(Block):
    timeout: Annotated[int, Config()] = 30  # Timeout in seconds

    method: Annotated[HTTPMethod, Config()] = HTTPMethod.GET
    url: Annotated[str, Config()] = ""
    headers: Annotated[dict[str, Any] | None, Config()] = None
    query_params: Annotated[dict[str, Any] | None, Config()] = None
    body: Annotated[dict[str, Any] | None, Config()] = None

    @step(output_name="response")
    async def make_request(
        self,
        request: Annotated[
            RequestObject,
            Metadata(
                description="Can accept request parameters (method, url, headers, query_params, and/or body). Any values not specified will use the Config values. Pass an empty dict ({}) to use all Config values"
            ),
        ],
    ) -> ResponseObject:
        # Helper to get the effective value from request or config
        def get_effective_value(attr: str):
            return getattr(request, attr) or getattr(self, attr)

        async with httpx.AsyncClient(timeout=self.timeout) as client:
            try:
                url = get_effective_value("url")
                if not url:
                    raise ValueError("URL is required")
                response = await client.request(
                    method=get_effective_value("method"),
                    url=url,
                    headers=get_effective_value("headers") or {},
                    params=get_effective_value("query_params") or {},
                    json=get_effective_value("body")
                    if get_effective_value("method") in ["POST", "PUT", "PATCH"]
                    else None,
                )

                response.raise_for_status()

                content_type = response.headers.get("content-type", "")
                body = None
                if "application/json" in content_type:
                    try:
                        body = response.json()
                    except ValueError:
                        # JSON decoding failed, leave body as None
                        pass

                return ResponseObject(
                    status_code=response.status_code,
                    headers=dict(response.headers),
                    text=response.text,
                    content=response.content,
                    body=body,
                )

            except httpx.RequestError as e:
                raise HTTPError(f"Network error occurred: {str(e)}")
            except httpx.HTTPStatusError as e:
                response_obj = ResponseObject(
                    status_code=e.response.status_code,
                    headers=dict(e.response.headers),
                    text=e.response.text,
                    content=e.response.content,
                    body=None,
                )
                raise HTTPError(
                    f"HTTP error occurred: {str(e)}",
                    e.response.status_code,
                    response_obj,
                )
            except Exception as e:
                raise HTTPError(f"Unexpected error occurred: {str(e)}")